/*
 * device.c
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 */


#include<config.h>
#include<types.h>
#include<lib.h>
#include<mm.h>
#include<interrupt.h>
#include<proc.h>
#include<fs.h>
#include<lock.h>
#include<device.h>
#include<except_i386.h>
#include<errno.h>
#include<lock.h>
#include<time.h>
#include<signal.h>
#include<test.h>
#include<debug.h>


//#define DEBUG_DEVICE 1

#ifdef DEBUG_DEVICE
	#define STATIC
	#define INLINE
#else
	#define STATIC static
	#define INLINE	inline
#endif


/******************************************************************************************************
 *
 * ǥХԤ
 *
 *******************************************************************************************************/


//================================== PUBLIC =============================================


void wait_intr(WAIT_INTR *wait, uint time, int procStat)
{
	int eflag;

	enterCli(&eflag);
	enter_spinlock(&wait->lock);
	if (wait->proc == NULL){
		wait->proc = get_current_task();

		exit_spinlock(&wait->lock);
		exitCli(&eflag);

		setWake(time);
		del_from_schedule(procStat, wait->isDel);
		wait_task();
		delWakeTimer(wait->proc->timer);
	}
	else{
		exit_spinlock(&wait->lock);
		exitCli(&eflag);
	}

	wait->proc = NULL;
	wait->isDel = PROC_SCHEDULE_DEL;
}


int wake_intr(WAIT_INTR *wait, int procStat)
{
	enter_spinlock(&wait->lock);
	if (wait->proc != NULL){
		exit_spinlock(&wait->lock);
		add_to_schedule(wait->proc, procStat, &wait->isDel);
		return 1;
	}
	else{
		wait->proc = WAIT_INTR_ACCESS;
		exit_spinlock(&wait->lock);
		return 0;
	}
}


/***********************************************************************************
 *
 * ֥åǥХ
 *
 ***********************************************************************************/


enum{
	/* Partition table */
	PARTITION_TBL_MAX		= 4,		/* 1κѡƥơ֥ */
	PARTITION_BEGIN			= 0x1be,	/* ѡƥơ֥볫ϥեå */
	PARTITION_MAGIC_ADDR	= 0x1fe,	/* ѡƥơ֥ޥåʥСեå */
	PARTITION_MAGIC_NUMBER	= 0xaa55,	/* ѡƥơ֥ޥåʥС */

	/* Partition type */
	PARTITION_TYPE_NULL		= 0,	/* NULL partition */
	PARTITION_TYPE_EXT		= 0x5,	/* ĥ */
};


/* Partition table */
typedef struct{
	uint8_t		active;		/* Boot active flag=0x80 */
	uint8_t		chsBeg[3];	/* CHS begin block */
	uint8_t		type;		/* Partition type */
	uint8_t		chsEnd[3];	/* CHS last block */
	uint32_t	lbaBeg;		/* LBA begin block */
	uint32_t	size;		/* Partition size by sector */
}PARTITON_TBL;


//================================== ѥ֥å ========================================


/*
 * ѡƥơ֥򸡺
 * return : NOERR
 *          ENODEV=Ѥ or ĥѡƥ 
 *          ENOENT=ĥѡƥ󥯤κǸã
 */
STATIC int readPartition(const DEV_INFO *devInfo,const int partNum,PARTITON_TBL *o_partTbl)
{
	char *sectBuf;
	PARTITON_TBL *partTbl;
	uint extBeg,extNext;
	int c_part;
	int i,rest;


	if (devInfo->open() < 0)
		return -EIO;

	/* MBRɤ */
	sectBuf = kmalloc(devInfo->sector_size);
	if (sectBuf == NULL)
		return -ENOMEM;
	rest = devInfo->read(sectBuf,1,0);
	if (rest < 0)
		goto ERR;

	/* ѡƥơ֥ǧ */
	if (*(uint16_t*)(sectBuf + PARTITION_MAGIC_ADDR) != PARTITION_MAGIC_NUMBER)
	{
		rest = -ENODEV;
		goto ERR;
	}
	partTbl = (PARTITON_TBL*)(sectBuf + PARTITION_BEGIN);

	/* ܥѡƥ򸡺 */
	if (partNum <= PARTITION_TBL_MAX)
	{
		c_part = partNum - 1;

		switch (partTbl[c_part].type)
		{
			/* ΰγĥѡƥϳʤ */
			case PARTITION_TYPE_EXT:
				/* ե륹롼NULLѡƥƱ */
			case PARTITION_TYPE_NULL:
				rest = -ENODEV;
				goto ERR;
			default:
				/* ѥѡƥơ֥ */
				o_partTbl->type		= partTbl[c_part].type;
				o_partTbl->lbaBeg	= partTbl[c_part].lbaBeg;
				o_partTbl->size		= partTbl[c_part].size;
		}
	}
	/* ĥѡƥ򸡺 */
	else
	{
		/*
		 * MBRγĥѡƥ򸡺
		 * out : c_part = ĥѡƥֹ
		 */
		for(i = 0;; ++i)
		{
			if (PARTITION_TBL_MAX <= i)
			{
				rest = -ENOENT;
				goto ERR;
			}
			if (partTbl[i].type == PARTITION_TYPE_EXT)
			{
				c_part = i;
				break;
			}
		}

		/*
		 * ĥѡƥ󥯤򤿤ɤ
		 * out : extNext = ĥΰϤޤ꤫Υ֥å sectBuf = ĥΰΥѡƥ󥻥
		 */
		extBeg = extNext = partTbl[c_part].lbaBeg;
		for(i = PARTITION_TBL_MAX; ;)
		{
			rest = devInfo->read(sectBuf,1,extNext);
			if (rest < 0)
				goto ERR;

			/* ѡƥȯ */
			if (++i == partNum)
				break;

			/* ĥѡƥ󥯤νλ */
			if (partTbl[1].type != PARTITION_TYPE_EXT)
			{
				rest = -ENOENT;
				goto ERR;
			}

			extNext = extBeg + partTbl[1].lbaBeg;
		}

		/* Ѥߥѡƥ */
		if (partTbl[0].type == PARTITION_TYPE_NULL)
		{
			rest = -ENODEV;
			goto ERR;
		}
		else
		{
			/* ѥѡƥơ֥ */
			o_partTbl->type		= partTbl[0].type;
			o_partTbl->lbaBeg	= partTbl[0].lbaBeg + extNext;
			o_partTbl->size		= partTbl[0].size;
		}
	}

	kfree(sectBuf);

	return NOERR;

ERR:
	kfree(sectBuf);

	return rest;
}


/***********************************************************************************
 *
 * ǥХե륷ƥ
 *
 * ¤
 *	inode¤Τǳ
 *		root inodeϣΩ
 *		̤inodeξ۴ĥꥹȤϢ
 *		ѺinodeꥹȤϢ
 ***********************************************************************************/


enum{
	MAX_DEVICE_NAME		= 8,	/* ǥХ̾ʸ */
	MAX_PART_NUM_DIGIT	= 4,	/* ֥åѡƥֹ */
	ROOT_INODE			= 0,	/* root inodeֹ */
};


/* ǥХinode */
typedef struct DEV_INODE{
	struct DEV_INODE	*next;
	struct DEV_INODE	*prev;
	DEV_INFO			*dev;					/* DEV_INFO¤Υݥ */
	uchar				 partNum;				/* ѡƥֹ */
	uchar				 partType;				/* ѡƥ󥿥 */
	short				 mode;					/* ե⡼ */
	int 				 refCount;				/* ȥ */
	uint				 blockBeg;				/* ǥǤΥѡƥλϤΥ֥å */
	uint				 blocks;				/* ѡƥ֥å */
	uint				 creatTime;				/*  */
	char				 name[MAX_DEVICE_NAME];	/* ǥХ̾ */
}DEV_INODE;


/* ǥХե륷ƥinodeơ֥ */
typedef struct{
	DEV_INODE	 tbl[MAX_DEVICE_OPEN];	/* inodeơ֥ */
	DEV_INODE	*top;					/* inodeơ֥󥯤κǽ */
	DEV_INODE	*free;					/* ̤inodeơ֥󥯤κǽ */
	uint		 lock;					/* ԥå */
}DEV_FS_TABLE;


//================================== ץ饤١ ======================================


static DEV_FS_TABLE devFs;


/*
 * root inode
 */
STATIC void setRootInode()
{
	/* inode */
	devFs.tbl[ROOT_INODE].partType	= 0;
	devFs.tbl[ROOT_INODE].partNum	= 0;
	devFs.tbl[ROOT_INODE].refCount	= 2;
	devFs.tbl[ROOT_INODE].dev		= NULL;
	devFs.tbl[ROOT_INODE].blockBeg	= 0;
	devFs.tbl[ROOT_INODE].blocks	= 0;
	devFs.tbl[ROOT_INODE].creatTime	= sys_time(NULL);
	devFs.tbl[ROOT_INODE].mode		= S_IFDIR | S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
}


/*
 * ǥХinodeݥ󥿤inodeֹ
 */
STATIC INLINE int getInodeNum(DEV_INODE *devInode)
{
	return ((uint)devInode - (uint)devFs.tbl) / sizeof(DEV_INODE);
}


/*
 * ѡƥֹʸѴ롣
 * ʸ"MAX_PART_NUM_DIGIT"ڤ롣
 */
STATIC void partNumToStr(int value,char *o_str)
{
	char str[MAX_PART_NUM_DIGIT + 1];
	int i;


	for (i = 0; i < MAX_PART_NUM_DIGIT; ++i)
	{
		str[i] = value % 10 + '0';
		if ((value /= 10) == 0)
			break;
	}
	o_str[i + 1] = '\0';
	while (i >= 0)
		*o_str++ = str[i--];
}


//================================== ѥ֥å ========================================


static DEV_INFO devInfoTop;


#ifdef DEBUG
STATIC int checkInode(int inode)
{
	if ((0 <= inode) && (inode < MAX_DEVICE_OPEN))
		return NOERR;
	else
	{
		printk("%s checkInode() Bad inode number:%d\n",__FILE__,inode);
		return -EINVAL;
	}
}
#endif


/*
 * inodeλȥ󥿤䤹
 */
STATIC void linkInode(int inode)
{
	ASSERT(0 < devFs.tbl[inode].refCount);
	devFs.tbl[inode].refCount += 1;
}


/*
 * inodeλȥ󥿤򣱸餹
 */
STATIC void unlinkInode(int inode)
{
	ASSERT(0 < devFs.tbl[inode].refCount);
	devFs.tbl[inode].refCount -= 1;
}


/*
 * inode
 * return : inode number or error number
 */
STATIC int getUnusedInode()
{
	DEV_INODE *free;
	
	
	/* inode뤫 */
	if (devFs.free == NULL)
		return -EMFILE;
	
	enter_spinlock(&devFs.lock);
	{
		free = devFs.free;

		/* Ǹζinode */
		if (free->next == free)
			devFs.free = NULL;
		else
		{
			/* inodeꥹȤФ */
			free->prev->next = free->next;
			free->next->prev = free->prev;
			devFs.free = free->next;
		}
		
		/* ѺѤinodeꥹȤˤĤʤ */
		free->next = devFs.top;
		devFs.top = free;
	}
	exit_spinlock(&devFs.lock);

	return getInodeNum(free);
}


/*
 * ǥХθ
 * return : inode or error number
 */
STATIC int lookup(const char *path)
{
	const char *devName;
	DEV_INODE *devInode;
	DEV_INODE *p;
	
	
	/*
	 * ѥθ
	 * out : devName = "path"ФǥХ̾
	 */
	for (devName = path;;)
	{
		/* "."".."ν */
		if (cmpPath(".",devName,1) == 0)
		{
			if (path[1] == '/')
				devName = &path[2];
			else
				devName = &path[1];
		}
		/* ".."̤б */
		else if (cmpPath("..",devName,2) == 0)
			return -ENOENT;
		/* root inode? */
		else if (cmpPath("\0",devName,0) == 0)
		{
			if (path[1] == '/')
				devName = &path[1];
			else
				return ROOT_INODE;
		}
		else
			break;
	}

	/*
	 * inode򸡺
	 * out : devInode = NULL or Ĥäinodeݥ
	 */
	for (p = devFs.top;; p = p->next)
	{
		if (p == NULL)
		{
			devInode = NULL;
			break;
		}
		if (strcmp(p->name,devName) == 0)
		{
			devInode = p;
			break;
		}
	}
	if (devInode == NULL)
		return -ENOENT;
	else
		return getInodeNum(devInode);
}


/*
 * ǥХinode롣
 * parameters : device name,begin block,blocks,partition type,partition number,FS
 * return : device inode or error number
 */
STATIC int setInode(const char *devName,uint blockBeg,uint size,int partType,int partNum,DEV_INFO *devInfo)
{
	char numStr[MAX_PART_NUM_DIGIT + 1] = {'\0'};
	char name[MAX_DEVICE_NAME + MAX_PART_NUM_DIGIT];
	int inode;


	/* ѡƥֹ礷ǥХ͡ʸγǧ */
	if (0 < partNum)
		partNumToStr(partNum,numStr);
	strcpy(name,devName);
	strcat(name,numStr);
	if (strlen(name) >= MAX_DEVICE_NAME)
		return -ENAMETOOLONG;
	
	/* ǤϿ뤫 */
	inode = lookup(name);

	/* inodeʤ */
	if (inode < 0)
	{

		/* inode */
		inode = getUnusedInode();
		if (inode < 0)
			return inode;

		/* inode */
		devFs.tbl[inode].partType = partType;
		devFs.tbl[inode].partNum = partNum;
		devFs.tbl[inode].refCount = 1;
		devFs.tbl[inode].dev = devInfo;
		devFs.tbl[inode].blockBeg = blockBeg;
		devFs.tbl[inode].blocks = size;
		devFs.tbl[inode].creatTime = sys_time(NULL);
		/* 饯ǥХ */
		if (devInfo->sector_size == 0)
			devFs.tbl[inode].mode = S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP;
		/* ֥åǥХ */
		else
			devFs.tbl[inode].mode = S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP;
		/* ǥХ̾ */
		strcpy(devFs.tbl[inode].name,name);
	}
	else
		return -EEXIST;

	return inode;
}


/*
 * Ϳ줿inodeμΥ󥯤inode֤
 * Ϳ줿inode-1ʤȥåץ󥯤inode֤
 * parameters : device inode
 * return : next device inode or ERR
 */
STATIC int getNextInode(int inode)
{
	if (inode == -1)
	{
		if (devFs.top == NULL)
			return ERR;
		return getInodeNum(devFs.top);
	}
	else
	{
		/* ̤ */
		if (devFs.tbl[inode].refCount == 0)
			return ERR;
		/* inode̵ */
		if (devFs.tbl[inode].next == NULL)
			return ERR;
		return getInodeNum(devFs.tbl[inode].next);
	}
}


/*
 * statꤹ
 */
STATIC int getStat(int inode,struct stat *o_stat)
{
	if (devFs.tbl[inode].refCount == 0)
		return -ENOENT;

	o_stat->st_dev			= inode;
	o_stat->st_ino			= 0;
	o_stat->st_mode			= devFs.tbl[inode].mode;
	o_stat->st_nlink		= devFs.tbl[inode].refCount;
	o_stat->st_uid			= 0;
	o_stat->st_gid			= 0;
	o_stat->st_size			= 0;
	o_stat->st_atime		= devFs.tbl[inode].creatTime;
	o_stat->st_mtime		= devFs.tbl[inode].creatTime;
	o_stat->st_ctime		= devFs.tbl[inode].creatTime;
	o_stat->st_crtpos		= 0;
	if (devFs.tbl[inode].dev == NULL)
		o_stat->st_blksize	= 0;
	else
		o_stat->st_blksize	= devFs.tbl[inode].dev->sector_size;
	o_stat->st_rdev			= inode;
	o_stat->st_flags		= 0;
	o_stat->st_blocks		= devFs.tbl[inode].blocks;

	return NOERR;
}


/*
 * DEV_INFO¤Υɥ쥹֤
 */
STATIC INLINE DEV_INFO *getDevInfo(int inode)
{
	return 	devFs.tbl[inode].dev;
}


/*
 * ֥å֤
 */
STATIC INLINE uint getBlocks(int inode)
{
	return 	devFs.tbl[inode].blocks;
}


/*
 * ֥åλϤޤ֤
 */
STATIC INLINE uint getBegBlock(int inode)
{
	return 	devFs.tbl[inode].blockBeg;
}


/*
 * ե⡼ɤ֤
 */
STATIC INLINE uint getMode(int inode)
{
	return 	devFs.tbl[inode].mode;
}


/*
 * ǥХ̾
 */
STATIC void getName(int inode,char *o_name)
{
	strcpy(o_name,devFs.tbl[inode].name);
}


/*
 * Init device filesystem.
 */
STATIC void initDevFs()
{
	int i;


	/* root inode */
	setRootInode();

	/* ٤Ƥinode̤ѥ󥯤˽۴ĥꥹȤ³ */
	devFs.free = &devFs.tbl[ROOT_INODE + 1];
	for(i = ROOT_INODE + 2;; ++i)
	{
		devFs.tbl[i].prev = &devFs.tbl[i - 1];
		devFs.tbl[i - 1].next = &devFs.tbl[i];
		if (i < MAX_DEVICE_OPEN - 1)
		{
			devFs.tbl[i].next = &devFs.tbl[i + 1];
			devFs.tbl[i + 1].prev = &devFs.tbl[i];
		}
		else
		{
			devFs.tbl[i].next = &devFs.tbl[ROOT_INODE + 1];
			devFs.tbl[ROOT_INODE + 1].prev = &devFs.tbl[i];
			break;
		}
	}
}


//=================================== Х =======================================


/*
 * Regist device.
 * parameters : Device infomation struct address
 * return : device inode or Error number
 */
int regist_device(DEV_INFO *devInfo)
{
	DEV_INFO *p;
	int rest;


	/* Name sizeγǧ  */
	if (MAX_DEVICE_NAME <= strlen(devInfo->name))
		return -ENAMETOOLONG;

	/* 󥯤³ */
	p = &devInfoTop;
	for (;p->next != NULL; p = p->next)
		switch (strcmp(devInfo->name,p->next->name))
		{
			case 0:
				return -EEXIST;
			case -1:
				goto EXIT;
		}
EXIT:
	devInfo->next = p->next;
	p->next = devInfo;

	/* ǥХե륷ƥϿ */
	rest = setInode(devInfo->name,devInfo->begin_blk,devInfo->last_blk - devInfo->begin_blk + 1,0,0,devInfo);
	if (rest < 0)
		return rest;

	return 0;
}


/*
 * ǥХμ
 * parameters : device inode,stat store buffer
 * return : 0 or error=-1
 */
int get_devstat(int inode,DEV_STAT *o_devStat)
{
	if (devFs.tbl[inode].refCount == 0)
		return -1;

	o_devStat->prt_type = devFs.tbl[inode].partType;
	o_devStat->prt_num = devFs.tbl[inode].partNum;
	o_devStat->sect_size = devFs.tbl[inode].dev->sector_size;
	o_devStat->all_sect = devFs.tbl[inode].blocks;
	o_devStat->name = devFs.tbl[inode].dev->name;

	return 0;
}


/*
 * Delete from regist table
 * parameters : name
 * return : 0,error -1
 */
int delete_device(const char *name)
{
	return 0;
}


/*
 * ֥åǥХѡƥǥХե륷ƥϿ
 * parmeters : ǥХ̵ֹ̾
 */
int registBlockPartition(const char *name)
{
	int inode;
	DEV_INFO *devInfo;
	PARTITON_TBL partTbl;
	int partNum;
	
	
	/* inode */
	inode = lookup(name);
	if (inode < 0)
		return inode;
	
	/* DEV_INFO¤Τ */
	devInfo = getDevInfo(inode);
	if (devInfo == NULL)
		return -ENODEV;
	
	/* ΰθ */
	for (partNum = 1; partNum <= PARTITION_TBL_MAX; ++partNum)
		if (readPartition(devInfo,partNum,&partTbl) == NOERR)
			setInode(devInfo->name,partTbl.lbaBeg,partTbl.size,partTbl.type,partNum,devInfo);

	/* ĥΰθ */
	for (partNum = PARTITION_TBL_MAX + 1;; ++partNum)
	{
		int rest;

		switch (rest = readPartition(devInfo,partNum,&partTbl))
		{
			case NOERR:
				setInode(devInfo->name,partTbl.lbaBeg,partTbl.size,partTbl.type,partNum,devInfo);
				break;
			case -ENODEV:
				continue;
			case -ENOENT:
				goto END;
			default:
				printk("%s registBlockPartition() unkown return value!:%d",__FILE__,rest);
		}
	}
END:

	return NOERR;
}


/************************************************************************************************
 *
 * ľžؿ
 *
 ************************************************************************************************/


//=================================== Х =======================================


/*
 * parameters : device inode number,buffer address,transfer sectors,begin sector
 * return : transfer sectors or error=-1
 */
int read_direct(int din,void *buf,size_t sectors,size_t begin)
{
	/* ǥ¤Υå */
	if (getBlocks(din) < begin)
		return -1;
	if (getBlocks(din) < begin + sectors)
		sectors = getBlocks(din) - begin;

	return getDevInfo(din)->read(buf,sectors,begin + getBegBlock(din));
}


/*
 * parameters : device inode number,buffer address,transfer sectors,begin sector
 * return : transfer sectors or error=-1
 */
int write_direct(int din,void *buf,size_t sectors,size_t begin)
{
	/* ǥ¤Υå */
	if (getBlocks(din) < begin)
		return -1;
	if (getBlocks(din) < begin + sectors)
		sectors = getBlocks(din) - begin;

	return getDevInfo(din)->write(buf,sectors,begin + getBegBlock(din));
}


/************************************************************************************************
 *
 * ǥХå
 *
 * ա
 *  ǥХϥ512ХȤˤƤΤǡ¾Υ
 *  ǥХоݤˤˤϡɬס
 *
 ************************************************************************************************/


enum{
	CACHE_NUM = 8,				/* IOå֥å */
	CACHE_BLOCK_SIZE = 0x1000,	/* IOå֥å */
	HASH_NUM = CACHE_NUM,		/* Number of cache hash table */
	CACHE_UNIT_SIZE = 512,		/* å֥åΥӥåȥޥåñ̥Хȥ */
};

typedef struct CACHE_TABLE{
	struct CACHE_TABLE			*hashNext;	// Hash next
	struct CACHE_TABLE			*hashPrev;
	struct CACHE_TABLE			*cacheNext;	// Cache chain next
	struct CACHE_TABLE			*cachePrev;
	struct CACHE_TABLE volatile	*writeNext;	// ٱ񤭹ߥꥹ
	struct CACHE_TABLE		 	*writePrev;
	char						*cache;		// Cache address
	uint						block;		// Device block index
	uchar						din;		// Device inode number
	uchar volatile				bitmap;		// ǡ¸ߥӥåȥޥå
	uchar volatile				writing;	// ٱԤե饰
	uchar volatile				busy;		// ե饰
	int							gate;		// ԥå
}CACHE_TABLE;


//================================== PRIVATE ============================================


static CACHE_TABLE cache_tbl[CACHE_NUM];			// Cache table
static CACHE_TABLE *cache_hash[HASH_NUM];			// Hash table for searching cache table
static CACHE_TABLE *cacheOldest = &cache_tbl[0];	// old chache block.nextֿ֥å
static int cacheLock;								// å󥯹ѥå
static PROC *writeTask;								// ٱ񤭹ѥ
static CACHE_TABLE delayTop;						// ٱ񤭹ߥꥹȤƬ
static int delayLock;								// å󥯹ѥå
static uchar bitSects[] = {							// ȤΥӥåȥޥå
	0,0x1,0x3,0x7,0xf,0x1f,0x3f,0x7f,0xff
};


// ǥХ饭åɤ߹ࡣ
// return : 0 or error number
STATIC int readDevice(CACHE_TABLE *cacheTbl, uchar bitmap)
{
	//ɤ߹ɬפɤ߹ޤƤʤӥåȤΤߣˤ
	uchar map = ~bitmap | cacheTbl->bitmap;
	DEV_INFO *devInfo = getDevInfo(cacheTbl->din);
	int start, last, allLast;
	int rest;

	for (start = 0, allLast = CACHE_BLOCK_SIZE / devInfo->sector_size; start < allLast; ++start){
		if ((map & (1 << start)) == 0){
			for (last = start + 1; last < allLast; ++last){
				if (map & (1 << last)){
					break;
				}
			}
			rest = devInfo->read(cacheTbl->cache + start * CACHE_UNIT_SIZE,
				last - start, start + cacheTbl->block + getBegBlock(cacheTbl->din));
			if (rest < 0){
				return rest;
			}
			start = last;
		}
	}

	cacheTbl->bitmap |= bitmap;
	
	return 0;
}


/*
 * å夫ǥХ˽᤹
 * return : 0 or error number
 */
STATIC int writeDevice(CACHE_TABLE *cacheTbl)
{
	DEV_INFO *devInfo = getDevInfo(cacheTbl->din);
	int start, last;
	int rest;

	if (cacheTbl->bitmap == 0xff){
		rest = devInfo->write(cacheTbl->cache, CACHE_BLOCK_SIZE / CACHE_UNIT_SIZE, cacheTbl->block + getBegBlock(cacheTbl->din));
		if (rest < 0){
			return rest;
		}
	}
	else{
		// սǥХ˽񤭹ࡣ
		for (start = 0; start < CACHE_BLOCK_SIZE / CACHE_UNIT_SIZE; ++start){
			if (cacheTbl->bitmap & (1 << start)){
				for (last = start + 1; last < CACHE_BLOCK_SIZE / CACHE_UNIT_SIZE; ++last){
					if ((cacheTbl->bitmap & (1 << last)) == 0){
						break;
					}
				}
				rest = devInfo->write(
					cacheTbl->cache + start * CACHE_UNIT_SIZE, last - start, cacheTbl->block + start + getBegBlock(cacheTbl->din));
				if (rest < 0){
					return rest;
				}
				start = last;
			}
		}
	}

	return 0;
}


/*
 * ǥХؤٱ񤭹ߥ
 */
STATIC void delayWriteTask()
{
	PROC **q;
	CACHE_TABLE *cacheTbl = NULL;

	// ƥץڤΥ
	for (q = &writeTask->parent->child; *q != writeTask; q = &(*q)->brother);
	*q = writeTask->brother;
	
	// λ륷ʥ򤹤٤ƥޥ
	setSigAllMask();

	// å
	enter_spinlock(&delayLock);
	for (;;){
		// 񤭹ԤФ
		if (delayTop.writeNext == &delayTop){
			if (cacheTbl != NULL){
				cacheTbl->writing = NO;
			}
			del_from_schedule(TASK_WAIT, PROC_SCHEDULE_DEL);

			// å
			exit_spinlock(&delayLock);

			wait_task();

			// å
			enter_spinlock(&delayLock);
		}
		else{
			cacheTbl = delayTop.writePrev;
			cacheTbl->writeNext->writePrev = cacheTbl->writePrev;
			cacheTbl->writePrev->writeNext = cacheTbl->writeNext;
			cacheTbl->writeNext = cacheTbl->writePrev = cacheTbl;

			// å
			exit_spinlock(&delayLock);

			if (writeDevice(cacheTbl) < 0){
				printk("Device error in delay write function!\n");
			}

			// å
			enter_spinlock(&delayLock);
			
			if (cacheTbl->writeNext == cacheTbl){
				cacheTbl->writing = NO;
			}
		}
	}
}


/*
 * ! δؿ"cacheLock"Ƕ礷ʤؿƤӽФ뤳
 * ٱ񤭹ߥꥹȤ˲ä롣
 */
STATIC void addWriteList(CACHE_TABLE *cacheTbl)
{
	enter_spinlock(&delayLock);
	{
		// Ǥ˥󥯤³Ƥ⤢Τǰöǽ򤹤
		cacheTbl->writeNext->writePrev = cacheTbl->writePrev;
		cacheTbl->writePrev->writeNext = cacheTbl->writeNext;
		
		cacheTbl->writeNext = delayTop.writeNext;
		cacheTbl->writePrev = &delayTop;
		delayTop.writeNext->writePrev = cacheTbl;
		delayTop.writeNext = cacheTbl;
		cacheTbl->writing = YES;

		// ٱ񤭹ߥ򵯤
		add_to_schedule(writeTask, TASK_WAIT, NULL);
	}
	exit_spinlock(&delayLock);

}


/*
 * ϥåͤ׻롣
 */
STATIC INLINE int calcHash(int value)
{
	return (value/8)%HASH_NUM;
}


/*
 * åơ֥뤫鳺ǥХ֥åõ
 * return : block addres or Failed=NULL
 */
STATIC CACHE_TABLE *searchCache(int din, size_t begin)
{
	int hash = calcHash(begin);
	CACHE_TABLE *cacheFind;

	for (cacheFind = cache_hash[hash]; cacheFind != NULL; cacheFind = cacheFind->hashNext){
		// åˤ
		if ((cacheFind->block == begin) && (cacheFind->din == din)){
			// 󥯤ֿ֤ˤ
			if (cacheFind == cacheOldest){
				cacheOldest = cacheOldest->cachePrev;
			}
			else{
				cacheFind->cacheNext->cachePrev = cacheFind->cachePrev;
				cacheFind->cachePrev->cacheNext = cacheFind->cacheNext;
				cacheFind->cacheNext = cacheOldest->cacheNext;
				cacheFind->cachePrev = cacheOldest;
				cacheOldest->cacheNext->cachePrev = cacheFind;
				cacheOldest->cacheNext = cacheFind;
			}

			return cacheFind;
		}
	}

	return NULL;
}


/*
 * ˥å֥åƤ롣
 * return : cache table
 */
STATIC CACHE_TABLE *getNewCache(int din, size_t begin)
{
	int hash = calcHash(begin);
	CACHE_TABLE *newCache;

	// ָŤå򲣼
	newCache = cacheOldest;
	cacheOldest = cacheOldest->cachePrev;

	// ϥå󥯤
	newCache->hashPrev->hashNext = newCache->hashNext;
	if (newCache->hashNext != NULL){
		newCache->hashNext->hashPrev = newCache->hashPrev;
	}

	// ϥå󥯤˲ä
	newCache->hashNext = cache_hash[hash];
	newCache->hashPrev = (CACHE_TABLE*)&cache_hash[hash];
	if (cache_hash[hash] != NULL){
		cache_hash[hash]->hashPrev = newCache;
	}
	cache_hash[hash] = newCache;

	return newCache;
}


/*
 * ٱ񤭹ߥεư
 */
STATIC void startDelayWriteTask()
{
	switch(sys_fork())
    {
		case -1:
			printk("Faild startDelayWriteTask\n");
			break;
		case 0:
			writeTask = get_current_task();
			delayWriteTask();
			// äƤʤ
	}
}


//================================== PROTECTED ==========================================


/*
 * ߤꥭå֥å򤹤٤ٱߤ롣
 */
STATIC void writeBackAllCache()
{
	// ٱ񤭹Ԥ
	for (;;){
		enter_spinlock(&delayLock);
		if (delayTop.writeNext != &delayTop){
			exit_spinlock(&delayLock);
			wait_task();
		}
		else{
			exit_spinlock(&delayLock);
			break;
		}
	}
}


/*
 * return : 0 or Error number
 */
STATIC int initCache()
{
	int i;

	// åơ֥ν
	for (i = 0; i < CACHE_NUM; ++i){
		memset(&cache_tbl[i], 0, sizeof(CACHE_TABLE));
		cache_tbl[i].cache = kmalloc(CACHE_BLOCK_SIZE);
		if (cache_tbl[i].cache == NULL){
			return -ENOMEM;
		}
		cache_tbl[i].cacheNext = &cache_tbl[i + 1];
		cache_tbl[i].cachePrev = &cache_tbl[i - 1];
		cache_tbl[i].hashPrev = &cache_tbl[i];
		cache_tbl[i].writeNext = cache_tbl[i].writePrev = &cache_tbl[i];
	}
	cache_tbl[CACHE_NUM - 1].cacheNext = &cache_tbl[0];
	cache_tbl[0].cachePrev = &cache_tbl[CACHE_NUM - 1];

	// ϥå󥯤ν
	for (i = 0; i < HASH_NUM; ++i){
		cache_hash[i] = NULL;
	}
	
	// ٱ񤭹ߥ󥯤ν
	delayTop.writeNext = delayTop.writePrev = &delayTop;
	
	return 0;
}


//================================== PUBLIC =============================================


/*
 *  ߥϥɥ餫θƤӽФԲ
 * return : transfer sectors or error number
 */
int read_cache(int din, void *buf, size_t sector_num, size_t begin)
{
	char *readBuf = (char*)buf;
	int sectSize = getDevInfo(din)->sector_size;	/* ǥХ */
	int cacheSects = CACHE_BLOCK_SIZE / sectSize;	/* å֥åˤϤ륻 */
	int restSects = sector_num;						/* Ĥž */
	int start, cacheStart, copySects;
	CACHE_TABLE *cacheTbl;

	ASSERT(0 <= restSects);

	/* ǥ¤Υå */
	if (getBlocks(din) < begin){
		return -EIO;
	}
	if (getBlocks(din) < begin + cacheSects){
		return -EIO;
	}
	start = begin % cacheSects;
	cacheStart = ROUNDDOWN(begin, cacheSects);
	copySects = (restSects <= cacheSects - start)? restSects : cacheSects - start;
	for(;;){
		enter_spinlock(&cacheLock);
		{
			// å򸡺
			if ((cacheTbl = searchCache(din,cacheStart)) != NULL){
				// åԤ
				while (cacheTbl->busy == YES){;
				}
			}
			else{
				cacheTbl = getNewCache(din,cacheStart);

				// åԤ
				while (cacheTbl->busy == YES){;
				}
				// ٱ񤭹Ԥ
				while (cacheTbl->writing == YES){;
				}

				cacheTbl->din = din;
				cacheTbl->block = cacheStart;
				cacheTbl->bitmap = 0;
			}

			// å
			cacheTbl->busy = YES;
		}
		exit_spinlock(&cacheLock);

		// å׵᥻ʤХǥХɤ߹
		if ((cacheTbl->bitmap & (bitSects[copySects] << start)) != (bitSects[copySects] << start)){
			if (readDevice(cacheTbl, bitSects[copySects] << start) < 0){
				// ---> åλ
				cacheTbl->busy = NO;
				return -EIO;
			}
		}

		// Хåե˥ԡ
		memcpy(readBuf, cacheTbl->cache + start * sectSize, copySects * sectSize);

		// åλ
		cacheTbl->busy = NO;

		if ((restSects -= copySects) == 0){
			break;
		}
		readBuf += copySects * sectSize;
		cacheStart += cacheSects;
		copySects = (restSects < cacheSects)? restSects : cacheSects;
		start = 0;
	}

	return sector_num;
}


/*
 *  ߥϥɥ餫θƤӽФԲ
 * return : transfer sectors or error number
 */
int write_cache(int din, void *wbuf, size_t sector_num, size_t begin)
{
	char *writeBuf = (char*)wbuf;
	int sectSize = getDevInfo(din)->sector_size;	/* ǥХ */
	int cacheSects = CACHE_BLOCK_SIZE / sectSize;	/* å֥åˤϤ륻 */
	int restSects = sector_num;						/* Ĥž */
	int start, cacheStart, copySects;
	CACHE_TABLE *cacheTbl;

	ASSERT(0 <= restSects);

	/* ǥ¤Υå */
	if (getBlocks(din) < begin){
		return -EIO;
	}
	if (getBlocks(din) < begin + cacheSects){
		return -EIO;
	}

	start = begin % cacheSects;
	cacheStart = ROUNDDOWN(begin, cacheSects);
	copySects = (restSects <= cacheSects - start)? restSects : cacheSects - start;
	for(;;){
		enter_spinlock(&cacheLock);
		{
			// å򸡺
			if ((cacheTbl = searchCache(din,cacheStart)) != NULL){
				// åԤ
				while (cacheTbl->busy == YES){;
				}
			}
			else{
				cacheTbl = getNewCache(din,cacheStart);

				// åԤ
				while (cacheTbl->busy == YES){;
				}
				// ٱ񤭹Ԥ
				while (cacheTbl->writing == YES){;
				}

				cacheTbl->din = din;
				cacheTbl->block = cacheStart;
				cacheTbl->bitmap = 0;
			}

			// å
			cacheTbl->busy = YES;
		}
		exit_spinlock(&cacheLock);

		// å˥ԡ
		memcpy(cacheTbl->cache + start * sectSize, writeBuf, copySects * sectSize);
		writeBuf += copySects * sectSize;
		cacheTbl->bitmap |= bitSects[copySects] << start;

		// ٱ񤭹
		addWriteList(cacheTbl);

		// åλ
		cacheTbl->busy = NO;

		if ((restSects -= copySects) == 0){
			break;
		}
		cacheStart += cacheSects;
		copySects = (restSects < cacheSects)? restSects : cacheSects;
		start = 0;
	}

	return sector_num;
}


int initBlockCache()
{
	/* IOå */
	initCache();

	/* ٱ񤭹ѥư롣 */
	startDelayWriteTask();

	return 0;
}


/************************************************************************************************
 *
 * ե륵ӥؿ
 *
 ************************************************************************************************/


/*
 * open()
 */
STATIC int open(const char *path,int dev,uint dir,int flag,uint *o_inode)
{
	int inode;


	/* ѥγǧ */
	inode = lookup(path);
	if (inode < 0)
		return inode;

	/* ǥХopen */
	if (inode != ROOT_INODE)
		if (getDevInfo(inode)->open() < 0)
			return -EIO;

	*o_inode = inode;

	return getMode(inode) & S_IFMT;
}


/*
 * close()
 */
STATIC int close(int din,uint inode)
{
#ifdef DEBUG
	if (checkInode(inode) < 0)
		return -EINVAL;
#endif
	/* ǥХclose */
	if (inode != ROOT_INODE)
		if (getDevInfo(inode)->close() < 0)
			return -EIO;

	return 0;
}


/*
 * read()
 */
STATIC int read(int din,uint inode,void *buf,size_t size,size_t begin)
{
	char *sectBuf = NULL;
	int sectSize;			/*  */
	int sectBeg;			/* žϤΥֹ */
	int restSize = size;	/* ĤžХȿ */
	int sects;				/*  */
	int bytes,restBytes;	/* Хȿ */
	union{
		void *pv;
		uint ui;
	}value;
	int rest;


#ifdef DEBUG
	if (checkInode(inode) < 0)
		return -EINVAL;
#endif
	switch (getMode(inode) & S_IFMT)
	{
		/* 饯ǥХ */
		case S_IFCHR:
			return getDevInfo(inode)->read(buf,size,0);

		/* ֥åǥХ */
		case S_IFBLK:
			sectSize = getDevInfo(inode)->sector_size;
			sectBeg = begin / sectSize;

			/* ǽΥ̤ž */
			if ((bytes = begin % sectSize) != 0)
			{
				if ((sectBuf = kmalloc(sectSize)) == NULL)
					return -ENOMEM;
				rest = read_cache(inode,sectBuf,1,sectBeg);
				if (rest != 1)
					return rest;
				restBytes = sectSize - bytes;
				if (size < restBytes)
				{
					memcpy(buf,sectBuf + bytes,size);
					kfree(sectBuf);
					return size;
				}

				memcpy(buf,sectBuf + bytes,restBytes);
				value.pv = buf;
				value.ui += restBytes;
				buf = value.pv;
				sectBeg += 1;
				restSize -= restBytes;
			}

			sects = restSize / sectSize;
			rest = read_cache(inode,buf,sects,sectBeg);
			if (rest < sects)
			{
				kfree(sectBuf);
				return rest;
			}

			/* Ĥ꤬̤ž */
			if (sects * sectSize < restSize)
			{
				bytes = sects * sectSize;
				if ((sectBuf == NULL) && ((sectBuf = kmalloc(sectSize)) == NULL))
					return -ENOMEM;
				if (read_cache(inode,sectBuf,1,sectBeg + sects) != 1)
				{
					kfree(sectBuf);
					return size - restSize + bytes;
				}
				memcpy((char*)buf + bytes,sectBuf,restSize - bytes);
			}

			kfree(sectBuf);
			
			return size;

		/* root directory */
		case S_IFDIR:
			return 0;
		
		default:
			printk("%s read() Bad file type!. Device file migt be destroied : inode=%d\n",__FILE__,inode);
			return -ENOENT;
	}
}

/*
 * write()
 */
STATIC int write(int din,uint inode,void *buf,size_t size,size_t begin)
{
	char *sectBuf = NULL;
	int sectSize;			/*  */
	int restSize = size;	/* ĤžХȿ */
	int sectBeg;			/* žϤΥֹ */
	int sects;				/*  */
	int bytes,restBytes;	/* Хȿ */
	union{
		void *pv;
		uint ui;
	}value;
	int rest;



#ifdef DEBUG
	if (checkInode(inode) < 0)
		return -EINVAL;
#endif
	switch (getMode(inode) & S_IFMT)
	{
		/* 饯ǥХ */
		case S_IFCHR:
			return getDevInfo(inode)->write(buf,size,0);

		/* ֥åǥХ */
		case S_IFBLK:
			sectSize = getDevInfo(inode)->sector_size;
			sectBeg = begin / sectSize;

			/* ǽΥ̤ž */
			if ((bytes = begin % sectSize) != 0)
			{
				if ((sectBuf = kmalloc(sectSize)) == NULL)
					return -ENOMEM;
				rest = read_cache(inode,sectBuf,1,sectBeg);
				if (rest != 1)
					return rest;
				if ((restBytes = sectSize - bytes) > size)
				{
					memcpy(sectBuf + bytes,buf,size);
					if (write_cache(inode,sectBuf,1,sectBeg) != 1)
						return 0;
					kfree(sectBuf);
					return size;
				}
				memcpy(sectBuf + bytes,buf,restBytes);
				if (write_cache(inode,sectBuf,1,sectBeg) != 1)
					return 0;
				value.pv = buf;
				value.ui += restBytes;
				buf = value.pv;
				sectBeg += 1;
				restSize -= restBytes;
			}

			sects = restSize / sectSize;
			rest = write_cache(inode,buf,sects,sectBeg);
			if (rest < sects)
			{
				kfree(sectBuf);
				return rest;
			}

			/* Ĥ꤬̤ž */
			if (sects * sectSize < restSize)
			{
				bytes = sects * sectSize;
				if ((sectBuf == NULL) && ((sectBuf = kmalloc(sectSize)) == NULL))
					return -ENOMEM;
				if (read_cache(inode,sectBuf,1,sectBeg + sects) != 1)
				{
					kfree(sectBuf);
					return size - restSize + bytes;
				}
				memcpy(sectBuf,(char*)buf + bytes,restSize - bytes);
				rest = write_cache(inode,sectBuf,1,sectBeg + sects);
				if (rest != 1)
				{
					kfree(sectBuf);
					return rest;
				}
			}

			kfree(sectBuf);

			return size;

		/* root directory */
		case S_IFDIR:
			return 0;
		
		default:
			printk("%s write() Bad file type!. Device file migt be destroied : inode=%d\n",__FILE__,inode);
			return -ENOENT;
	}
}


/*
 * opendir()
 */
STATIC int opendir(const char *path,int din,uint dirInode,uint *o_inode,uint *o_block,int *o_index)
{
	int inode;


	/* ѥθ */
	inode = lookup(path);
	if (inode < 0)
		return inode;
	
	*o_inode = inode;
	*o_block = 0;
	*o_index = -1;

	return 0;
}


STATIC int readdir(int din,uint inode,uint *m_block,int *m_index,char *o_name)
{
	int nextInode;


	if (inode != ROOT_INODE)
		return -ENOENT;

	/* Υȥ꡼򸡺 */
	nextInode = getNextInode(*m_index);
	if (nextInode == ERR)
		return -ENOENT;
	
	*m_index = nextInode;

	/* ̾ */
	getName(nextInode,o_name);

	return 0;
}


STATIC int stat(int din,uint inode,struct stat *o_stat)
{
#ifdef DEBUG
	if (checkInode(inode) < 0)
		return -EINVAL;
#endif

	return getStat(inode,o_stat);
}


/*
 * parameters : device inode,inode,Command,Parameter struct pointer
 */
STATIC int ioctl(int din,uint inode,int cmd,...)
{
	uint *arg = (uint*)&cmd + 1;


#ifdef DEBUG
	if (checkInode(inode) < 0)
		return -EINVAL;
#endif
	if (inode == ROOT_INODE)
		return -ENOENT;

	return getDevInfo(inode)->ioctl(cmd,*arg,*(arg + 1));
}


STATIC int mount(int din,uint *inode)
{
	return -EOPNOTSUPP;
}


STATIC int umount(int din,uint inode)
{
	return -EOPNOTSUPP;
}


STATIC int rename(int din,uint old_dinode,const char *old_path,uint new_dinode,const char *new_path)
{
	return -EOPNOTSUPP;
}


STATIC int creat(const char *path,int din,uint dir_inode,int mode,uint *inode)
{
	return -EOPNOTSUPP;
}


STATIC int mkdir(const char *path,int din,uint dir_inode,int mode)
{
	return -EOPNOTSUPP;
}

STATIC int delete(const char *path,int din,uint dir_inode,int type)
{
	return -EOPNOTSUPP;
}

STATIC int chattr(int din,uint inode,uint mode,uint uid,uint gid,uint atime,uint mtime)
{
	return -EOPNOTSUPP;
}

STATIC int link(int din,uint oldinode,uint newind,const char *newpath)
{
	return -EOPNOTSUPP;
}

STATIC int symlink(const char *path,int din,uint dir_inode,const char *link_path)
{
	return -EOPNOTSUPP;
}

STATIC int select(uint inode,int flag,int polling)
{
#ifdef DEBUG
	if (checkInode(inode) < 0)
		return -EINVAL;
#endif
	if (inode == ROOT_INODE)
		return -ENOENT;

	return getDevInfo(inode)->select(flag,polling);
}

STATIC int statfs(int din,struct statfs *buf)
{
	return -EOPNOTSUPP;
}


static FS devFsFunc = {
	DEV_FS_NAME,
	mount,
	umount,
	open,
	close,
	read,
	write,
	ioctl,
	rename,
	creat,
	opendir,
	mkdir,
	stat,
	delete,
	readdir,
	chattr,
	link,
	symlink,
	select,
	statfs,
};


/*
 * å夫ǥХ˥ǡ᤹
 */
int sys_sync()
{
	writeBackAllCache();

	return 0;
}


/*
 * եǥץü̾롣
 */
int sys_ttyname(int fdnum,char *buf)
{
	F_DSC *fd;
	DEV_INFO *dev;


	if ((fd = getFd(fdnum)) == NULL)
		return -EBADF;

	/* DEV_INFOΥ饹ȥ֥åʥС0Υ饯ǥХǧ롣 */
	dev = getDevInfo(fd->vent->inode);
	if (dev->sector_size != 0)
		return -ENOTTY;

	memcpy(buf,dev->name,strlen(dev->name) + 1);

	return 0;
}


/************************************************************************************************
 *
 * 
 *
 ************************************************************************************************/


//=================================== Х =======================================


int initDeviceFs()
{
	/* ǥХե륷ƥν */
	initDevFs();
	
	return 0;
}


int mountDeviceFs()
{
	int rest;


	/* ե륷ƥϿ롣 */
	rest = regist_fs(&devFsFunc);
	if (rest < 0)
		return rest;

	/* ǥХե륷ƥޥȤ롣 */
	rest = mount_dev_fs(&devFsFunc);
	if (rest < 0)
		return rest;

	return 0;
}


/****************************************************************************/
void test_device()
{
	printk("buf=%x\n",*(uint*)0x158f58);
}
/*****************************************************************************/
