/*
 * fs.c
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 * ۥե륷ƥࡣ
 * ե륷ƥ
 */


#include<types.h>
#include<config.h>
#include<limits.h>
#include<lib.h>
#include<mm.h>
#include<proc.h>
#include<lock.h>
#include<errno.h>
#include<interrupt.h>
#include<time.h>
#include<signal.h>
#include<device.h>
#include<net/net.h>
#include<share/utime.h>
#include<fs.h>
#include<test.h>
#include<debug.h>


//#define DEBUG_FS 1

#ifdef DEBUG_FS
	#define STATIC
	#define INLINE
#else
	#define STATIC	static
	#define INLINE	inline
#endif

//	#define CALL_PRINT()	PRINT_ERR("")
	#define CALL_PRINT()

/***************************************************************************************
 *
 * ۥե륷ƥ
 *
 *  ޥȻȥǥ쥯ȥѹե륪ץե¹Իɲä롣
 *  ȥ꡼ϡ"."+".."+̾ͽ¤٤롣
 *  ȥ꡼󥯤κǸϺǽΥȥ꡼Ϣ뤹롣
 *  ǥ쥯ȥϥޥȤ줿ΰʳϥǥХ֥åǤ϶
 *  ƥॳǸüȥ꡼ϡΥƥॳ˺ʤ褦
 *  󥿤䤷ΥƥॳκǸǥ󥿤򸺤餹
 *
 ***************************************************************************************/

enum{
	VFS_MAX_NAME_SIZE = PATH_MAX,		/* ۥե륷ƥκ̾ */
};


typedef struct{
	char *path;
	VENTRY *vent;
}SEARCH_PARAM;


/* PUBLIC 롼ȥȥ꡼ */
static char _rootEnt[sizeof(VENTRY)+1];
static VENTRY *rootEnt = (VENTRY*)&_rootEnt;

/* PUBLIC devǥ쥯ȥꥨȥ꡼ */
static char _devEnt[sizeof(VENTRY)+1];
static VENTRY *devEnt=(VENTRY*)&_devEnt;

/* PUBLIC ۥե륷ƥॹԥåȡ */
static int spinLock=0;


/*
 * PRIVATE
 * ȥ꡼󥯤饨ȥ꡼򸡺롣
 * Υȥ꡼ɥ쥹ݥ󥿡ˤϸĤäȥ꡼ޤϡ
 * ǸΥȥ꡼ɥ쥹Хåե롣
 * parameters : ѥɥ쥹,ǥ쥯ȥꥨȥ꡼ɥ쥹ݥ󥿡,entry buffer
 * return : next path or failed=NULL
 */
STATIC char *searchEntry(char *path,VENTRY *entry,VENTRY **pEntry)
{
	int rest;
	VENTRY *last;



	if (cmpPath(".",path,1) == 0)
	{
		*pEntry = entry;
		if (path[1] == '/')
			return &path[2];
		else
			return &path[1];
	}
	if (cmpPath("..",path,2) == 0)
	{
		*pEntry = entry->parent;
		if (path[2] == '/')
			return &path[3];
		else
			return &path[2];
	}

	if ((last = entry = entry->child) == NULL)
	{
		*pEntry = NULL;
		return NULL;
	}

	switch (rest = cmpPath(entry->name,path,entry->len))
	{
		case -1:
			for (entry = entry->next; entry != last; entry = entry->next)
				if ((rest = cmpPath(entry->name,path,entry->len)) != -1)
					break;
			break;
		case 1:
			*pEntry = NULL;
			return NULL;
	}

	*pEntry = entry;

	if (rest == 0)
	{
		for (path += entry->len; *path == '/'; ++path);
		return path;
	}

	return NULL;
}


/*
 * PRIVATE
 * ȥ꡼롣
 * parameters : entry addres
 * return : 0 or error number
 */
STATIC int delEntry(VENTRY *entry)
{
	char name[VFS_MAX_NAME_SIZE],*pname;
	VENTRY *parent;


	if (entry->refCount > 0)
		return -EBUSY;
	if (entry->child != NULL)
		return -EBUSY;

	parent = entry->parent;

	if (entry->next == entry)
		parent->child = NULL;
	else
	{
		if (parent->child == entry)
			parent->child = entry->next;
		entry->prev->next = entry->next;
		entry->next->prev = entry->prev;
	}

	/* եκ */
	if (entry->link-- <= 0)
	{
		name[VFS_MAX_NAME_SIZE-1] = '\0';
		pname = &name[VFS_MAX_NAME_SIZE-1];
		memcpy(pname -= entry->len,entry->name,entry->len);
		for (parent = entry->parent; parent->inode == 0; parent = parent->parent)
		{
			if (pname + parent->len + 1 < name)
				goto EXIT;
			*--pname = '/';
			memcpy(pname -= parent->len,parent->name,parent->len);
		}
		entry->fs->delete(pname,parent->din,parent->inode,S_IFREG);
	}
EXIT:
	kfree(entry->wait);
	kfree(entry);

	return 0;
}


/*
 * PRIVATE
 * ȥ꡼̤äƺ롣
 * parameters : ȥ꡼
 */
STATIC void delUpEntry(VENTRY *entry)
{
	VENTRY *parent;

	for(;;)
	{
		parent = entry->parent;
		if (delEntry(entry) != 0)
			break;
		entry=parent;
	}
}


/*
 * PUBLIC
 * ѥ̾Υȥ꡼򸡺롣
 * ǽΰˤϸƤӽФͤͿ졢ޥȤ줿ǥ쥯ȥ꤬Ф롣
 * ΰϥѥκǽΥȥ꡼롣
 * parameters : start search parmeters struct,last search parmeters struct
 * return : 0 or error number
 */
STATIC int searchPath(SEARCH_PARAM *start,SEARCH_PARAM *end)
{
	char *pPath,*retPath;
	VENTRY *entry,*pEntry;


	entry = start->vent;
	for (pPath = start->path; *pPath != '\0';)
	{
		if (entry->type != S_IFDIR)
			return -ENOTDIR;
		if ((retPath = searchEntry(pPath,entry,&pEntry)) == NULL)
			break;
		pPath = retPath;
		entry = pEntry;
		if (entry->mount != NULL)
		{
			start->path = pPath;
			start->vent = entry->mount;
			entry = entry->mount;
		}
	}

	end->path = pPath;
	end->vent = entry;

	return 0;
}


/*
 * PUBLIC
 * ѥƬΡ̾Υȥ꡼󥯤ɲä롣
 * parameters : path pointer,parent entry,file type
 * return : new entry address
 */
STATIC VENTRY *addEntry(char *path,VENTRY *entry,int type)
{
	int len;
	VENTRY *next_ent,*new_ent;


	/* ȥ꡼򸡺 */
	if (searchEntry(path,entry,&next_ent) != NULL)
		return NULL;

	/* ȥ꡼κ */
	len = pathLen(path);
	new_ent = kmalloc(sizeof(VENTRY) + len);
	if (new_ent == NULL)
		return NULL;
	memset(new_ent,0,sizeof(VENTRY));
	new_ent->fs = entry->fs;
	new_ent->din = entry->din;
	new_ent->type = type;
	new_ent->parent = entry;
	new_ent->len = len;
	new_ent->link = 1;
	memcpy(new_ent->name,path,len);

	/* 󥯤ɲä롣 */
	if (next_ent != NULL)
	{
		new_ent->prev = next_ent->prev;
		new_ent->next = next_ent;
		next_ent->prev->next = new_ent;
		next_ent->prev = new_ent;
	}
	else if (entry->child != NULL)
	{
		next_ent = entry->child;
		new_ent->prev = next_ent->prev;
		new_ent->next = next_ent;
		next_ent->prev->next = new_ent;
		next_ent->prev = new_ent;
		entry->child = new_ent;
	}
	else
	{
		new_ent->prev = new_ent;
		new_ent->next = new_ent;
		entry->child = new_ent;
	}

	return new_ent;
}


/*
 * PUBLIC
 * ȥ꡼λȥ󥿤1䤹
 * parameters : ۥȥ꡼
 */
STATIC INLINE void linkVentry(VENTRY *vent)
{
	vent->refCount += 1;
	vent->link += 1;
}


/*
 * PUBLIC
 * ȥ꡼λȥ󥿤򸺤餷0ʤ̥Ρɥȥ꡼̤äƺ롣
 * parameters : ۥȥ꡼
 */
STATIC INLINE int unlinkVentry(VENTRY *vent)
{
	int cnt;


	if ((cnt = --vent->refCount) <= 0)
	{
		delUpEntry(vent);
		return 0;
	}

	return cnt;
}


/*
 * PUBLIC
 * ȥ꡼Υ󥯥Ȥ򸺤餹
 */
STATIC INLINE void unlinkVentryLink(VENTRY *vent)
{
	--vent->link;
}


/*
 * PUBLIC
 * ȥ꡼ѥɲä롣
 * ա pathκǽ餬NULLʸǤʤȡ
 * parameters : path,ۥȥ꡼
 * return : last entry or error number
 */
STATIC int mkEntry(char *path,VENTRY *vent)
{
	for(;;++path)
	{
		vent = addEntry(path,vent,S_IFDIR);
		if (vent == NULL)
			return -ENOENT;
		path+=vent->len;

		if(*path=='\0')break;
	}
	vent->refCount=1;
	vent->link=1;

	return (int)vent;
}


/*
 * PUBLIC
 * ȥ꡼˿ȥ꡼ޥȤ롣
 * parameters : ޥȥȥ꡼,filesystem,device inode,root directory block
 * return : 0 or error number
 */
STATIC int mountEntry(VENTRY *entry,FS *fs,int din,uint block)
{
	VENTRY *mount_ent;


	/* Ǥ˥ޥȤƤ뤫 */
	if (entry->mount != NULL)
		return -EBUSY;

	/* ޥȥȥ꡼κ */
	if ((mount_ent = kmalloc(sizeof(VENTRY) + entry->len)) == NULL)
		return -ENOMEM;
	memset(mount_ent,0,sizeof(VENTRY));
	mount_ent->next		= entry->next;
	mount_ent->prev		= entry->prev;
	mount_ent->inode	= block;
	mount_ent->parent	= entry->parent;
	mount_ent->fs		= fs;
	mount_ent->din		= din;
	mount_ent->type		= S_IFDIR;
	mount_ent->len		= entry->len;
	memcpy(mount_ent->name,entry->name,entry->len);

	linkVentry(mount_ent);

	/* ޥȥȥ꡼˥ޥȤ롣 */
	linkVentry(entry);
	entry->mount = mount_ent;

	return 0;
}


/*
 * PUBLIC
 * ޥȡ
 * parameters : ޥȥȥ꡼
 * return : 0 or error number
 */
STATIC int umountEntry(VENTRY *entry)
{
	int rest;
	FS *fs;
	VENTRY *mount_ent;


	if((mount_ent=entry->mount)==NULL)return -ENOENT;
	if(mount_ent->refCount>1)return -EBUSY;
	if(mount_ent->child!=mount_ent->child->next->next)return -EBUSY;

	/* ޥȥȥ꡼κ */
	fs=mount_ent->fs;
	mount_ent->refCount=0;
	if((rest=delEntry(mount_ent))<0)return rest;
	if((rest=fs->umount(mount_ent->din,mount_ent->inode))<0)return rest;
	--entry->refCount;

	/* ޥȥȥ꡼ʤ */
	unlinkVentry(entry);

	return 0;
}


/*
 * PUBLIC
 * ѥ롼ȤޤäƥХåե˥ԡ롣
 * parameters : buffer,buffer size,ۥȥ꡼
 * return : write size or error=-1
 */
STATIC int copyCurrentPath(char *buf,size_t size,VENTRY *vent)
{
	int len;


	if(vent==rootEnt)
	{
		buf[0]='/';
		return 1;
	}
	else
	{
		if((len=copyCurrentPath(buf,size,vent->parent))==-1)return -1;
		if(len+vent->len+1>size)return -1;
		memcpy(&buf[len],vent->name,vent->len);
		buf[len+vent->len]='/';

		return len+vent->len+1;
	}
}


/*
 * PUBLIC
 * ۥȥ꡼롢ǳɬס
 * parameters : start search parmeters struct,last search parmeters struct
 * return : 0 or error number
 */
STATIC int getVentObj(SEARCH_PARAM *start,SEARCH_PARAM *end)
{
	int rest;


	enter_spinlock(&spinLock);
	{
		if ((rest = searchPath(start,end)) == 0)
			end->vent->refCount += 1;
	}
	exit_spinlock(&spinLock);

	return rest;
}


/*
 * PUBLIC
 * ۥȥ꡼ζꤹ롣
 */
STATIC void setVentObj(VENTRY *vent)
{
	enter_spinlock(&spinLock);
	{
		vent->refCount+=1;
	}
	exit_spinlock(&spinLock);
}


/*
 * PUBLIC
 * ۥȥ꡼γ
 */
STATIC void releaseVentObj(VENTRY *vent)
{
	enter_spinlock(&spinLock);
	{
		vent->refCount-=1;
	}
	exit_spinlock(&spinLock);
}


/*
 * PUBLIC
 * struct statfs˳Ǽ롣
 * return : 0 or error number
 */
STATIC int getFsstat(VENTRY *vent,struct statfs *buf)
{
	int rest;
	DEV_STAT devstat;


	if((rest=vent->fs->statfs(vent->din,buf))<0)
		return rest;
	get_devstat(vent->din,&devstat);

	if(devstat.prt_num>0)
		sprintk(buf->f_mntfromname,"/dev/%s%d",devstat.name,devstat.prt_num);
	else
		sprintk(buf->f_mntfromname,"/dev/%s",devstat.name,devstat);
	copyCurrentPath(buf->f_mntonname,MNAMELEN,vent);

	return 0;
}


/*
 * PRIVATE
 * ǥ쥯ȥޥȥȥ꡼θ
 * parameters : VENTRY,struct statfs,struct statfsΰ,struct statfs̵ե饰
 */
STATIC int searchMountDir(VENTRY *vent,struct statfs *buf,int buf_cnt,int flag)
{
	int rest,count;
	VENTRY *vent_begin;


	if(buf_cnt<=0)return 0;
	vent_begin=vent;
	count=buf_cnt;

	do
	{
		/* ޥȤꡣ */
		if(vent->mount!=NULL)
		{
			while(vent->mount!=NULL)vent=vent->mount;
			if(flag)
				if(getFsstat(vent,buf)==0)
				{
					buf+=1;
					count-=1;
				}	
		}

		if(vent->child!=NULL)
		{
			rest=searchMountDir(vent->child,buf,count,flag);
			buf+=rest;
			count-=rest;
		}
	}
	while((vent=vent->next)!=vent_begin);

	return buf_cnt-count;
}


/*
 * PUBLIC
 * ޥȥȥ꡼򸡺롣
 * parameters : struct statfs,buf count
 * return : struct statfsسǼ or ޥȿstruct statfs=NULL
 */
STATIC int searchMount(struct statfs *buf,int buf_cnt)
{
	int rest,flag;


	if(buf_cnt<=0)return 0;

	flag=(buf==NULL)?0:1;
	
	/* rootե륷ƥࡣ */
	if(flag==1)
		getFsstat(rootEnt,buf);

	rest=searchMountDir(rootEnt->child,buf+1,buf_cnt-1,flag);

	return rest+1;
}


/************************************************************************************
 *
 * ե륷ƥ
 *
 ************************************************************************************/

/* PRIVATE */
STATIC FS fsInfoTop;


/*
 * PUBLIC
 * Search file system.
 * parameters : file system name
 * return : FS address or NULL
 */
STATIC FS *search_fs(const char *name)
{
	FS *p;


	for(p=fsInfoTop.next;p!=NULL;p=p->next)
		switch(strcmp(name,p->name))
		{
			case 0:
				return p;
			case -1:
				return NULL;
		}

	return NULL;
}


/*
 * GLOBAL
 * Regist filesystem.
 * parameters : Filesystem struct
 * return : 0 or Error=-1
 */
int regist_fs(FS *info)
{
	FS *p;


	p=&fsInfoTop;
	for(;p->next!=NULL;p=p->next)
		switch(strcmp(info->name,p->next->name))
		{
			case 0:
				return -EEXIST;
			case -1:
				goto EXIT;
		}
EXIT:
	info->next=p->next;
	p->next=info;

	return 0;
}


/************************************************************************************
 *
 * ץΥե
 *
 ************************************************************************************/


//================================== PRIVATE ============================================


/*
 * Get empty fd number.
 * ահ"min""OPEN_MAX"ʾǤʤȡ
 * parameter : FILE_STRUCT,fd number
 * return : empty number or error number
 */
STATIC INLINE int getEmptyFdNum(FILE_STRUCT *fstruct, const int min)
{
	int num = min;

	for (num = min; fstruct->fd[num] != NULL;){
		if (OPEN_MAX <= ++num){
			num = 0;
		}
		if (num == min){
			return -ENFILE;
		}
	}

	return num;
}


/*
 * եǥץ鳺ե롣
 * parameters : F_DSC
 */
STATIC void delFromFd(F_DSC *fd)
{
	char path[VFS_MAX_NAME_SIZE];
	int len;


	/* 롼ȤޤǤΥѥ롣 */
	enter_spinlock(&spinLock);
	{
		len = copyCurrentPath(path,VFS_MAX_NAME_SIZE,fd->vent);
		if (len >= VFS_MAX_NAME_SIZE)
			return;
	}
	exit_spinlock(&spinLock);

	/*  */
	rootEnt->fs->delete(path,rootEnt->din,rootEnt->inode,S_IFREG);
}


//================================== PROTECTED ==========================================


/*
 * Get current directory entry.
 * return current directory entry
 */
STATIC INLINE VENTRY *getCurrentDir()
{
	return ((FILE_STRUCT*)get_current_task()->file_struct)->current_dir;
}


/*
 * Set current directory.
 */
STATIC INLINE void setCurrentDir(VENTRY *vent)
{
	FILE_STRUCT *fstruct;


	fstruct=get_current_task()->file_struct;
	unlinkVentry(fstruct->current_dir);
	fstruct->current_dir=vent;
}


/*
 * Get file descriptor flag.
 * parameters : file descriptor number,flag
 * return : flag or error number
 */
STATIC INLINE int getFdFlag(int num)
{
	FILE_STRUCT *fstruct;


	if((uint)num>=OPEN_MAX)return -EBADF;
	fstruct=get_current_task()->file_struct;
	if(fstruct->fd[num]==NULL)return -EBADF;

	return fstruct->flag[num];
}


/*
 * Set file descriptor flag.
 * parameters : file descriptor number,flag
 * return : 0 or error number
 */
STATIC INLINE int setFdFlag(int num,int flag)
{
	FILE_STRUCT *fstruct;


	if((uint)num>=OPEN_MAX)return -EBADF;
	fstruct=get_current_task()->file_struct;
	if(fstruct->fd[num]==NULL)return -EBADF;
	fstruct->flag[num]=flag;

	return 0;
}


/*
 * Get file open mode flag.
 * parameters : file descriptor number
 * returm : open mode flag
 */
STATIC INLINE int getOpenFlag(int num)
{
	FILE_STRUCT *fstruct;


	if((uint)num>=OPEN_MAX)return -EBADF;
	fstruct=get_current_task()->file_struct;
	if(fstruct->fd[num]==NULL)return -EBADF;

	return fstruct->fd[num]->accessMode;
}


/*
 * Set file open mode flag.
 * parameters : file descriptor number
 * returm : open mode flag
 */
STATIC INLINE int setOpenFlag(int num,int flag)
{
	FILE_STRUCT *fstruct;


	if((uint)num>=OPEN_MAX)return -EBADF;
	fstruct=get_current_task()->file_struct;
	if(fstruct->fd[num]==NULL)return -EBADF;

	fstruct->fd[num]->accessMode&=~(O_APPEND|O_NONBLOCK);
	fstruct->fd[num]->accessMode|=flag&(O_APPEND|O_NONBLOCK);

	return 0;
}


/*
 * Copy file descriptor.
 * parameters : ԡ FD number,ԡǾ FD number
 * return : file descriptor number or error number
 */
STATIC int copyFd(int num1,int num2)
{
	FILE_STRUCT *fstruct = get_current_task()->file_struct;

	// ǥץֹ
	if (OPEN_MAX <= (uint)num2){
		num2 = 0;
	}
	num2 = getEmptyFdNum(fstruct,num2);

	// եǥץ򥳥ԡ
	fstruct->fd[num2] = fstruct->fd[num1];
	++fstruct->fd_count;
	++fstruct->fd[num2]->refCount;
	fstruct->flag[num2] = 0;			// Clear close-on-exec flag.
	linkVentry(fstruct->fd[num2]->vent);

	return num2;
}


/*
 * file descriptorλȥ󥿤򸺤餹
 * parameters : FILE_STRUCT,fd number
 * return : 0 or error number
 */
STATIC int unlinkFd(FILE_STRUCT *fstruct, int num)
{
	int rest = 0;
	int din;
	uint inode;
	F_DSC *fd;
	FS *fs;

	fd = fstruct->fd[num];
	--fd->refCount;
	if (fd->form == FD_FORM_NETSOKET){
		if (fd->refCount <= 0){
			closeSocket((SOCKET*)fd);
		}
	}
	else{
		fs = fd->vent->fs;
		din = fd->vent->din;
		inode = fd->vent->inode;
		if (unlinkVentry(fd->vent) == 0){
			rest = fs->close(din,inode);
		}

		if (fd->refCount <= 0){
			// õե饰ǧ
			if (fd->accessMode & O_DEL){
				delFromFd(fd);
			}
			kfree(fd);
		}
	}

	fstruct->fd[num] = NULL;
	fstruct->next_fd = num;
	--fstruct->fd_count;

	return rest;
}


//================================== PUBLIC =============================================


/*
 * fork˥եطι¤Τƥץ饳ԡ롣
 * parameters : parent file structure
 * return : child file structure or error=NULL
 */
FILE_STRUCT *copyFileStruct(FILE_STRUCT *fstruct)
{
	FILE_STRUCT *cp_struct;
	int i,cnt;

	cp_struct = kmalloc(sizeof(FILE_STRUCT));
	if (cp_struct == NULL){
		return NULL;
	}
	memcpy(cp_struct, fstruct, sizeof(FILE_STRUCT));

	// եǥץλȥ󥿤1䤹
	for (cnt = fstruct->fd_count, i = 0; 0 < cnt; ++i){
		if (cp_struct->fd[i] != NULL){
			++cp_struct->fd[i]->refCount;
			if (cp_struct->fd[i]->form == FD_FORM_LOCALFILE){
				linkVentry(cp_struct->fd[i]->vent);
			}

			--cnt;
		}
	}

	return cp_struct;
}


/*
 * exit()Υեطι¤Τ롣
 * parameters : FILE_STRUCT address
 * return : 0 or error=-1
 */
void releaseFileStruct(FILE_STRUCT *fstruct)
{
	int i;

	for (i = 0; i < OPEN_MAX; ++i){
		if (fstruct->fd[i] != NULL){
			unlinkFd(fstruct, i);
			if (fstruct->fd_count <= 0){
				break;
			}
		}
	}

	kfree(fstruct);
}


/*
 * execΥե빽¤Τν
 */
void initExecFs()
{
	FILE_STRUCT *fstruct = get_current_task()->file_struct;
	int i, cnt;

	// Close file descriptor.
	for (cnt = fstruct->fd_count, i = 0; 0 < cnt; ++i){
		if (fstruct->fd[i] != NULL){
			if (fstruct->flag[i] & FD_CLOEXEC){
				unlinkFd(fstruct,i);
			}
			
			--cnt;
		}
	}
}


/*
 * եǥץinode֤
 * parameters : file descriptor
 * return : inode block or error=-1
 */
int get_inode(int fd)
{
	F_DSC *fdsc;


	if(fd>OPEN_MAX)return -1;
	if((fdsc=((FILE_STRUCT*)(get_current_task()->file_struct))->fd[fd])==NULL)return -1;

	return fdsc->vent->inode;
}


/*
 * ɥץѥե빽¤Τν
 * parameters : process
 * return : 0 or error number
 */
int init_file_struct(PROC *proc)
{
	FILE_STRUCT *fstruct;


	if((fstruct=kmalloc(sizeof(FILE_STRUCT)))==NULL)return -ENOMEM;
	memset(fstruct,0,sizeof(FILE_STRUCT));
	fstruct->current_dir=rootEnt;
	proc->file_struct=fstruct;

	return 0;
}


/*
 * եǥץϿ롣
 * parameters : ventry,file format,access mode,fd flag,F_DSC buffer
 * return : file descriptor number
 */
int setFd(VENTRY *vent, int form, int amode, int flag, F_DSC *fd)
{
	int num;
	FILE_STRUCT *fstruct = get_current_task()->file_struct;

	// եǥץν
	fd->form = form;
	fd->accessMode = amode;
	fd->offset = 0;
	fd->refCount = 1;
	fd->vent = vent;
	num = getEmptyFdNum(fstruct, fstruct->next_fd);
	fstruct->next_fd = (num + 1 < OPEN_MAX)? num + 1 : 0;
	++fstruct->fd_count;
	fstruct->fd[num] = fd;
	fstruct->flag[num] = flag;

	return num;
}


/*
 * File descriptor롣
 * parameters : file descriptor number
 * return : file descriptor or NULL
 */
F_DSC *getFd(int num)
{
	if (OPEN_MAX <= (uint)num){
		return NULL;
	}
	return ((FILE_STRUCT*)get_current_task()->file_struct)->fd[num];
}


/*
 * ץե礫
 * return : FALSE or TRUE
 */
int isFileOpenMax()
{
	FILE_STRUCT *fstruct = get_current_task()->file_struct;

	return  (OPEN_MAX <= fstruct->fd_count)? TRUE : FALSE;
}


/************************************************************************************
 *
 * ե륷ƥ
 *
 ************************************************************************************/

/*
STATIC int lookup(const char *path, const int accessMode, uint *o_inodeNum)
{
	SEARCH_PARAM start,end;
	int mode = 0;
	
	// ⡼ɤ3ӥåȤΥե⡼ɤѴ
	if (accessMode & O_RDONLY){
		mode |= S_IROTH;
	}
	if (accessMode & O_WRONLY){
		mode |= S_IWOTH;
	}
	if (accessMode & O_EXCL){
		mode |= S_IXOTH;
	}
	
	// ۥե륷ƥफ
}
*/

/************************************************************************************
 *
 * System call interface
 *
 ************************************************************************************/


/*
 * PUBLIC
 * pathȥåץǥ쥯ȥ롣
 * parameters : ѥɥ쥹ݥ
 * return : Ȳۥǥ쥯ȥꥨȥ꡼
 */
STATIC INLINE VENTRY *getTopVdir(const char *path,char **pPath)
{
	char *p = (char*)path;
	
	
	if(*p == '/')
	{
		while(*++p == '/');
		*pPath = p;
		return rootEnt;
	}
	else
	{
		*pPath = p;
		return getCurrentDir();
	}
}


/*
 * GLOBAL
 * ¹ԥե륪ץ̾Υץδάؿ
 * parameters : path,open file struct
 * return : 0 or error number
 */
int exec_open(const char *path,EXEC_FD *fd)
{
	uint inode;
	SEARCH_PARAM start,end;
/*	VENTRY *vent;*/
	int rest;


	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* եõơۥȥ꡼롣 */
	start.vent = getTopVdir(path,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;

	if ((*end.path == '\0') && (end.vent->type == S_IFDIR))
	{
		rest = -EISDIR;
		goto END;
	}

	if(*end.path == '\0')
	{
/*		linkVentry(end.vent);*/		/* ۥեؤΥ󥯤α */
		fd->din = end.vent->din;
		fd->inode = end.vent->inode;
		fd->fs = end.vent->fs;
	}
	else
	{
		rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,0,&inode);
		if (rest < 0)
			goto END;

		/* ۥեκα */
/*		enter_spinlock(&spinLock);
		{
			rest=mkEntry(end.path,end.vent);
		}
		exit_spinlock(&spinLock);

		if(rest<0)
		{
			start.vent->fs->close(start.vent->din,start.vent->inode);
			goto END;
		}
*/
		fd->din = start.vent->din;
		fd->inode = inode;
		fd->fs = start.vent->fs;
	}

	rest = 0;
END:
	releaseVentObj(end.vent);

	return rest;
}


/*
 * GLOBAL
 * ¹ԥեɤ߹ߡ"sys_read"δάؿ
 * parameters : open file struct,buffer,read size
 * return : read byte or error=-1
 */
int exec_read(EXEC_FD *fd, void *buf, size_t size)
{
	return fd->fs->read(fd->din, fd->inode, buf, size, fd->offset);
}


/*
 * GLOBAL
 * ¹ԥեΥ
 */
int exec_close(EXEC_FD *fd)
{
	return fd->fs->close(fd->din,fd->inode);
}


/*
 * PUBLIC
 * ޥȥǥХ
 * parameters : device name,filesystem number
 * return : device inode or error number
 */
STATIC int mountDevice(const char *dev_path)
{
	enum{DEV_NAME_SIZE=5};		/* "/dev/"Υ */

	int rest;
	uint inode;


	/* ǥХץ */
	if (cmpPathNull("/dev",dev_path) == NULL)
		return -EINVAL;

	rest = devEnt->mount->fs->open(dev_path+DEV_NAME_SIZE,0,0,0,&inode);
	if (rest < 0)
		return rest;

	return inode;
}

int sys_mount(const char *dev_path,const char *fs_name,const char *path)
{
	uint pInode;
	int mount_din;
	FS *fs;
	SEARCH_PARAM start,end;
	VENTRY *vent;
	int rest;


	CALL_PRINT();
	
	/* ѥĹθ */
	if (PATH_MAX <= strlen(dev_path))
		return -ENAMETOOLONG;
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* Search file system. */
	if((fs=search_fs(fs_name))==NULL)return -ENODEV;

	/* ǥХץ */
	if((rest=mountDevice(dev_path))<0)return rest;
	mount_din=rest;

	/* ޥȴؿƤӽФ */
	rest = fs->mount(mount_din,&pInode);
	if (rest < 0)
		goto ERR;

	/* ۥե륷ƥΥǥ쥯ȥ򸡺롣 */
	start.vent = getTopVdir(path,&start.path);
	if((rest=getVentObj(&start,&end))<0)
		goto ERR;

	if (*end.path != '\0')
	{
		uint inode;
		rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,0,&inode);
		if (rest  < 0)
			goto ERR2;
		start.vent->fs->close(start.vent->din,inode);
		if ((rest = mkEntry(end.path,end.vent)) < 0)
			goto ERR2;
		vent = (VENTRY*)rest;
	}
	else vent = end.vent;

	/* ޥȤ벾ۥǥ쥯ȥ롣 */
	if ((rest = mountEntry(vent,fs,mount_din,pInode)) < 0)
		goto ERR2;

	releaseVentObj(end.vent);

	return 0;
ERR2:
	releaseVentObj(end.vent);
ERR:
	fs->umount(mount_din,pInode);

	return rest;
}


int sys_umount(const char *path)
{
	int rest;
	SEARCH_PARAM start,end;


	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* ǥ쥯ȥõ */
	start.vent = getTopVdir(path,&start.path);
	rest = getVentObj(&start,&end);
	if (rest < 0)
		return rest;

	if ((*end.path != '\0') || (end.vent->type != S_IFDIR))
	{
		rest=-ENOENT;
		goto END;
	}

	/* ޥȡ */
	rest = umountEntry(end.vent);
END:
	releaseVentObj(end.vent);

	return rest;
}


int sys_lseek(int fdNum, int offset, int flag)
{
	int rest;
	FSTAT fstat;
	F_DSC *fd;
	VENTRY *vent;

	CALL_PRINT();

	fd = getFd(fdNum);
	if (fd == NULL)
		return -EBADF;
	vent = fd->vent;
	vent->fs->stat(vent->din,vent->inode,&fstat);
	switch(flag)
	{
		case SEEK_CUR:
			rest = fd->offset + offset;
			break;
		case SEEK_SET:
			rest = offset;
			break;
		case SEEK_END:
			rest = fstat.st_size - offset;
		default:
			return -EINVAL;
	}

	if (rest < 0)
		return -EINVAL;

	return fd->offset = rest;
}


int sys_creat(const char *path,int flag,int mode)
{
	int rest;
	uint inode;
	SEARCH_PARAM start,end;
	VENTRY *vent;
	F_DSC *fd;


	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	if(*path=='\0')
		return -ENOENT;

	/* ץΥå */
	if (isFileOpenMax() == TRUE){
		return -EMFILE;
	}

	/* եõơۥȥ꡼롣 */
	start.vent = getTopVdir(path,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;

	if (*end.path == '\0')
	{
		rest = -EEXIST;
		goto END;
	}

	if (flag & O_FIFO)
		mode |= S_IFIFO;
	else
		mode |=  S_IFREG | S_IRUSR | S_IWUSR;
	rest = start.vent->fs->creat(start.path,start.vent->din,start.vent->inode,mode,&inode);
	if (rest < 0)
		goto END;

	enter_spinlock(&spinLock);
	{
		rest = mkEntry(end.path,end.vent);
	}
	exit_spinlock(&spinLock);
	if (rest < 0)
	{
		start.vent->fs->close(start.vent->din,start.vent->inode);
		goto END;
	}

	vent = (VENTRY*)rest;
	vent->inode = inode;
	if (flag & O_FIFO)
	{
		if ((vent->wait = kmalloc(sizeof(WAIT_QUEUE))) == NULL)
			goto END;
		memset(vent->wait,0,sizeof(WAIT_QUEUE));
		vent->type=S_IFIFO;
	}
	else
		vent->type = S_IFREG;

	/* եǥץν */
	if((fd=kmalloc(sizeof(F_DSC)))==NULL)
	{
		rest=-ENOMEM;
		goto END;
	}
	rest=setFd(vent,FD_FORM_LOCALFILE,flag,FD_CLOEXEC,fd);
END:
	releaseVentObj(end.vent);

	return rest;
}


/*
 * return : file descriptor number or error number
 */
int sys_open(const char *path, int flag, int mode)
{
	uint inode;
	int rest, type;
	SEARCH_PARAM start, end;
	VENTRY *vent;
	F_DSC *fd;

	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path)){
		return -ENAMETOOLONG;
	}
	if ((flag & (O_CREAT | O_EXCL)) == (O_CREAT | O_EXCL)){
		return sys_creat(path,flag,mode);
	}
	/* ץΥå */
	if (isFileOpenMax() == TRUE){
		return -EMFILE;
	}

	/* եõơۥȥ꡼롣 */
	start.vent = getTopVdir(path,&start.path);
	rest = getVentObj(&start,&end);
	if (rest < 0){
		return rest;
	}

	if (*end.path == '\0'){
		if (end.vent->type == S_IFDIR){
			/* ǥ쥯ȥɤ߹ߤΤ */
			if (flag != O_RDONLY){
				rest = -EISDIR;
				goto END;
			}

			if (end.vent->refCount == 0){
				rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,0,&inode);
				if (rest < 0){
					goto END;
				}
				end.vent->inode = inode;
			}
		}
		vent = end.vent;
		if (flag & O_TRUNC){
			rest = vent->fs->ioctl(vent->din,vent->inode,I_TRUNC);
			if (rest < 0){
				goto END;
			}
		}
		linkVentry(vent);
	}
	else{
		rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,flag & O_TRUNC,&inode);
		if ((rest == -ENOENT) && (flag & O_CREAT)){
			mode |= S_IFREG | S_IRUSR | S_IWUSR;
			rest = start.vent->fs->creat(start.path,start.vent->din,start.vent->inode,mode,&inode);
		}
		if (rest < 0){
			goto END;
		}
		type = rest;

		/* ǥ쥯ȥɤ߹ߤΤ */
		if (type == S_IFDIR){
			if (flag != O_RDONLY){
				rest = -EISDIR;
				goto END;
			}
		}

		enter_spinlock(&spinLock);
		{
			rest = mkEntry(end.path,end.vent);
		}
		exit_spinlock(&spinLock);
		if (rest < 0){
			start.vent->fs->close(start.vent->din,inode);
			goto END;
		}

		vent = (VENTRY*)rest;
		vent->inode = inode;
		if (flag & O_FIFO){
			vent->wait = kmalloc(sizeof(WAIT_QUEUE));
			if (vent->wait == NULL){
				rest = -ENOMEM;
				goto END;
			}
			memset(vent->wait,0,sizeof(WAIT_QUEUE));
			vent->type = S_IFIFO;
		}
		else{
			vent->type = type;
		}
	}

	/* եǥץϿ롣 */
 	if ((fd = kmalloc(sizeof(F_DSC))) == NULL){
		rest = -ENOMEM;
		goto END;
	}
	rest = setFd(vent, FD_FORM_LOCALFILE, flag, FD_CLOEXEC, fd);
	if (flag & O_APPEND){
		sys_lseek(rest,0,SEEK_END);
	}
END:
	releaseVentObj(end.vent);

	return rest;
}


/*
 * parameters : file descriptor number
 * return : 0 or error number
 */
int sys_close(int fdNum)
{
	F_DSC *fd;
	FILE_STRUCT *fstruct;

	CALL_PRINT();

	if ((fd = getFd(fdNum)) == NULL)
		return -EBADF;
	fstruct = (FILE_STRUCT*)get_current_task()->file_struct;

	return unlinkFd(fstruct,fdNum);
}


/*
 * parameters : file descriptor number,baffer,read bytes
 * return : read bytes or error=-1
 */
int sys_read(int fdNum, void *buf, size_t size)
{
	F_DSC *fd;
	VENTRY *vent;
	int rest;

	CALL_PRINT();

	/* Хåեå */
	if (checkMem(buf, size) == -1){
		return -EFAULT;
	}

	if ((fd = getFd(fdNum)) == NULL){
		return -EBADF;
	}

	/* ⡼ɤΥå */
	if ((fd->accessMode & (O_RDONLY | O_RDWR)) == 0){
		return -EBADF;
	}

	/* readؿƤӽФ */
	vent = fd->vent;
	if (vent->type == S_IFIFO){
		waitSemaphore(vent->wait);
		{
			rest = vent->fs->read(vent->din, vent->inode, buf, size, 0);
		}
		wakeSemaphore(vent->wait);
	}
	else{
		rest = vent->fs->read(vent->din, vent->inode, buf, size, fd->offset);
		if (0 < rest){
			fd->offset += rest;
		}
	}

	return rest;
}


int sys_write(int fdNum, void *buf, size_t size)
{
	F_DSC *fd;
	VENTRY *vent;
	int rest;

	CALL_PRINT();

	/* Хåեå */
	if (checkMem(buf, size) == -1){
		return -EFAULT;
	}

	if ((fd = getFd(fdNum)) == NULL){
		return -EBADF;
	}

	/* ⡼ɤΥå */
	if ((fd->accessMode & (O_WRONLY | O_RDWR)) == 0){
		return -EBADF;
	}

	/* writeؿƤӽФ */
	vent = fd->vent;
	if (vent->type == S_IFIFO){
		waitSemaphore(vent->wait);
		{
			rest=vent->fs->write(vent->din,vent->inode,buf,size,0);
		}
		wakeSemaphore(vent->wait);
	}
	else{
		rest = vent->fs->write(vent->din, vent->inode, buf, size, fd->offset);
		if (0 < rest){
			fd->offset += rest;
		}
	}

	return rest;
}


int sys_mkdir(const char *path,int mode)
{
	int rest;
	SEARCH_PARAM start,end;


	CALL_PRINT();

	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	start.vent = getTopVdir(path,&start.path);
	if((rest=getVentObj(&start,&end))<0)return rest;
	if(*end.path=='\0')
	{
		rest=-EEXIST;
		goto END;
	}

	rest=start.vent->fs->mkdir(start.path,start.vent->din,start.vent->inode,mode);
END:
	releaseVentObj(end.vent);

	return rest;
}


int sys_unlink(const char *path)
{
	int rest;
	SEARCH_PARAM start,end;


	CALL_PRINT();

	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	start.vent = getTopVdir(path,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;

	if (*end.path != '\0'){
		rest = start.vent->fs->delete(start.path,start.vent->din,start.vent->inode,S_IFREG);
	}
	else {
		unlinkVentryLink(end.vent);
		rest = 0;
	}

	releaseVentObj(end.vent);

	return rest;
}


int sys_rmdir(const char *path)
{
	int rest;
	SEARCH_PARAM start,end;


	CALL_PRINT();

	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	start.vent = getTopVdir(path,&start.path);
	if((rest=getVentObj(&start,&end))<0)
		return rest;

	if(*end.path!='\0')
		rest=start.vent->fs->delete(start.path,start.vent->din,start.vent->inode,S_IFDIR);
	else
		rest=-EBUSY;

	releaseVentObj(end.vent);

	return rest;
}


/*
 * return : directory stream number or error number
 */
int sys_opendir(const char *path)
{
	uint block,inode;
	int index;
	SEARCH_PARAM start,end;
	VENTRY *vent;
	DIR_STRAEM *dir;
	int rest;


	CALL_PRINT();

	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* եõ */
	start.vent = getTopVdir(path,&start.path);
	rest = getVentObj(&start,&end);
	if (rest < 0)
		return rest;

	if(*end.path == '\0')
	{
		if(end.vent->type != S_IFDIR)
		{
			rest = -ENOTDIR;
			goto END;
		}
		linkVentry(end.vent);
		vent = end.vent;
		rest = start.vent->fs->opendir(end.path,vent->din,vent->inode,&inode,&block,&index);
		if (rest < 0)
			goto END;
	}
	else
	{
		if(end.vent->type == S_IFDIR)
		{
			rest = start.vent->fs->opendir(start.path,start.vent->din,start.vent->inode,&inode,&block,&index);
			if (rest < 0)
				goto END;

			/* ۥȥ꡼κ */
			enter_spinlock(&spinLock);
			{
				rest=mkEntry(end.path,end.vent);
			}
			exit_spinlock(&spinLock);
			if(rest<=0)
			{
				start.vent->fs->close(start.vent->din,start.vent->inode);
				goto END;
			}
			vent = (VENTRY*)rest;
			vent->inode = inode;
			vent->type = S_IFDIR;
		}
		else
		{
			rest = -ENOENT;
			goto END;
		}
	}

	/* ǥ쥯ȥꥹȥ꡼κ */
	dir = kmalloc(sizeof(DIR_STRAEM));
	if (dir == NULL)
	{
		rest=-ENOMEM;
		goto END;
	}
	rest = setFd(vent,FD_FORM_DIRSTREAM,0,FD_CLOEXEC,(F_DSC*)dir);
	dir->block = block;
	dir->index = index;
END:
	releaseVentObj(end.vent);

	return rest;
}


int sys_readdir(int dir_num,char *name)
{
	DIR_STRAEM *dir;
	VENTRY *vent;


	CALL_PRINT();

	if ((dir = (DIR_STRAEM*)getFd(dir_num)) == NULL)
		return -EBADF;
	if (dir->fd.form != FD_FORM_DIRSTREAM)
		return -EBADF;

	vent = dir->fd.vent;

	return vent->fs->readdir(vent->din,vent->inode,&dir->block,&dir->index,name);
}


int sys_rewinddir(int dir_num)
{
	DIR_STRAEM *dir;


	CALL_PRINT();

	if((dir=(DIR_STRAEM*)getFd(dir_num))==NULL)return -EBADF;
	if(dir->fd.form!=FD_FORM_DIRSTREAM)return -EBADF;
	dir->index=0;

	return 0;
}


/*
 * old pathnew pathƱ졢ͭå롣
 */
int sys_rename(const char *old_path,const char *new_path)
{
	int rest;
	SEARCH_PARAM old_start,old_end;
	SEARCH_PARAM new_start,new_end;


	CALL_PRINT();

	if(strcmp(old_path,new_path)==0)return 0;

	/* եõ */
	old_start.vent = getTopVdir(old_path,&old_start.path);
	if((rest=getVentObj(&old_start,&old_end))<0)
		return rest;

	/* եõ */
	new_start.vent = getTopVdir(new_path,&new_start.path);
	if((rest=getVentObj(&new_start,&new_end))<0)
		goto END;

	if(old_start.vent->din==new_start.vent->din)
	{
		if((*old_end.path!='\0')&&(*new_end.path!='\0'))
			rest=old_start.vent->fs->
				rename(old_start.vent->din,old_start.vent->inode,old_start.path,new_start.vent->inode,new_start.path);
		else rest=-EBUSY;
	}
	else rest=-EXDEV;

	releaseVentObj(new_end.vent);
END:
	releaseVentObj(old_end.vent);

	return rest;
}


int sys_chdir(const char *path)
{
	uint inode;
	int rest = 0;
	SEARCH_PARAM start,end;
	VENTRY *vent;
	int dummy1;
	uint dummy2;

	
	CALL_PRINT();

	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* ǥ쥯ȥõ */
	start.vent = getTopVdir(path,&start.path);
	rest = getVentObj(&start,&end);
	if (rest < 0)
		return rest;

	if (end.vent->type == S_IFDIR)
	{
		if (*end.path == '\0')
		{
			if (end.vent->refCount == 1)
			{
				rest = start.vent->fs->opendir(start.path,start.vent->din,start.vent->inode,&inode,&dummy1,&dummy2);
				if (rest < 0)
					goto ERR;
				end.vent->inode = inode;
			}
			linkVentry(end.vent);
			vent = end.vent;
		}
		else
		{
			rest = start.vent->fs->opendir(start.path,start.vent->din,start.vent->inode,&inode,&dummy1,&dummy2);
			if (rest >= 0)
			{
				enter_spinlock(&spinLock);
				{
					rest = mkEntry(end.path,end.vent);
				}
				exit_spinlock(&spinLock);

				if (rest <= 0)
				{
					start.vent->fs->close(start.vent->din,start.vent->inode);
					goto ERR;
				}
				vent = (VENTRY*)rest;
				vent->inode = inode;
				vent->type = S_IFDIR;
			}
			else goto ERR;
		}
	}
	else
	{
		rest = -ENOTDIR;
		goto ERR;
	}

	releaseVentObj(end.vent);

	/* ե빽¤Τ˿ȥǥ쥯ȥϿ */
	setCurrentDir(vent);

	return 0;
ERR:
	releaseVentObj(end.vent);

	return rest;
}


int sys_fchdir(int fdNum)
{
	F_DSC *fd;


	CALL_PRINT();

	fd = getFd(fdNum);
	if (fd == NULL)
		return -EBADF;
	if (fd->vent->type != S_IFDIR)return -ENOTDIR;

	linkVentry(fd->vent);
	
	/* ե빽¤Τ˿ȥǥ쥯ȥϿ */
	setCurrentDir(fd->vent);

	return 0;
}


int sys_getcwd(char *buf,size_t size)
{
	int len;
	VENTRY *vent;


	CALL_PRINT();

	if (size == 0)
		return -EINVAL;

	/* Хåեå */
	if (checkMem(buf,size) == -1)
		return -EFAULT;

	vent=getCurrentDir();
	if((len=copyCurrentPath(buf,size,vent))==-1)return -ERANGE;
	if(len==1)buf[1]='\0';
	else buf[len-1]='\0';

	return 0;
}


/*
 * pathNULLʤfile descriptorõ
 * parameters : file descriptor,path,stat buffer
 */
int sys_stat(int fdNum,const char *path,struct stat *o_stat)
{
	uint inode;
	SEARCH_PARAM start,end;
	F_DSC *fd;
	int rest;

	CALL_PRINT();

	if (checkMem(o_stat,sizeof(struct stat)) == -1){
		return -EFAULT;
	}

	if (path == NULL){
		if ((fd = getFd(fdNum)) == NULL){
			return -EBADF;
		}
		o_stat->st_crtpos = fd->offset;

		return fd->vent->fs->stat(fd->vent->din,fd->vent->inode,o_stat);
	}
	else{
		/* ѥĹθ */
		if (PATH_MAX <= strlen(path)){
			return -ENAMETOOLONG;
		}

		o_stat->st_crtpos = 0;

		start.vent = getTopVdir(path,&start.path);
		rest = getVentObj(&start,&end);
		if (rest < 0){
			return rest;
		}

		if ((*end.path == '\0') && (end.vent->inode != 0)){
			rest = end.vent->fs->stat(end.vent->din,end.vent->inode,o_stat);
		}
		else{
			rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,0,&inode);
			if (rest < 0){
				goto END;
			}
			rest = start.vent->fs->stat(start.vent->din,inode,o_stat);
			start.vent->fs->close(start.vent->din,inode);
		}
	}
END:
	releaseVentObj(end.vent);

/************** ̤ **************************/
	o_stat->st_flags = 0;
/*****************************************************/

	return rest;
}


/*
 * PRIVATE
 * ۥե륷ƥ򸡺ƥޥȤƤե륷ƥ򥢥ޥȤ롣
 * parameters : vertial entry
 * return : process or Failed=NULL
 */
STATIC void umountAllFs(VENTRY *ventry)
{
	VENTRY *vent=ventry;


	do
	{
		if(vent->child!=NULL)umountAllFs(vent->child);

		/* umouont file system. */
		if (vent->mount != NULL)
			vent->mount->fs->umount(vent->mount->din,vent->mount->inode);
	}while ((vent = vent->next) != ventry);
}


/*
 * 롼ȥե륷ƥ򥢥ޥȤ롣
 */
int sys_umount_root()
{
	CALL_PRINT();

	umountAllFs(rootEnt);
	rootEnt->fs->umount(rootEnt->din,rootEnt->inode);

	return 0;
}


int sys_fcntl(int fdNum,int cmd,uint arg)
{
	CALL_PRINT();

	switch(cmd)
	{
		case F_DUPFD:
		{
			if (isFileOpenMax() == TRUE){
				return -EMFILE;
			}
			if (getFd(fdNum) == NULL){			/* File descriptorΥå */
				return -EBADF;
			}
			return copyFd(fdNum,arg);			/* Copy File descriptor. */
		}
		case F_GETFD:
			return getFdFlag(fdNum);
		case F_SETFD:
			return setFdFlag(fdNum,arg);
		case F_GETFL:
			return getOpenFlag(fdNum);
		case F_SETFL:
			return setOpenFlag(fdNum,arg);
		case F_GETLK:
		case F_SETLK:
		case F_SETLKW:
		case F_GETOWN:
		case F_SETOWN:
		default:
			return -ENOTSUP;
	}
}


STATIC int chattr(FS *fs,int din,uint inode,int cmd,uint attr)
{
	int rest;
	PROC *proc;
	struct stat stat;


	/* Change attribute. */
	proc=get_current_task();
	if((rest=fs->stat(din,inode,&stat))<0)return rest;
	switch(cmd)
	{
		case CHATTR_MODE:
			if((proc->uid==0)||(proc->uid==stat.st_uid))
				rest=fs->chattr(din,inode,attr,stat.st_uid,stat.st_gid,stat.st_atime,stat.st_mtime);
			else rest=-EPERM;
			break;
		case CHATTR_UID:
			if(proc->uid==0)
				rest=fs->chattr(din,inode,stat.st_mode,attr,stat.st_gid,stat.st_atime,stat.st_mtime);
			else rest=-EPERM;
			break;
		case CHATTR_GID:
			if(((proc->uid==0)||(proc->uid==stat.st_uid))&&(attr==proc->gid))
				rest=fs->chattr(din,inode,stat.st_mode,stat.st_gid,attr,stat.st_atime,stat.st_mtime);
			else rest=-EPERM;
			break;
		case CHATTR_UTIME:
			if((proc->uid==0)||(proc->uid==stat.st_uid))
			{
				struct utimbuf *ut=(struct utimbuf*)attr;
				uint time;

				if(ut==NULL)
				{
					time=sys_time(NULL);
					rest=fs->chattr(din,inode,stat.st_mode,stat.st_gid,attr,time,time);
				}
				else
					rest=fs->chattr(din,inode,stat.st_mode,stat.st_gid,attr,ut->actime,ut->modtime);
			}
			else rest=-EPERM;
			break;
	}

	return rest;
}


int sys_chattr(const char *path,int cmd,uint attr)
{
	uint inode;
	int rest;
	SEARCH_PARAM start,end;


	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* Open inode. */
	start.vent = getTopVdir(path,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;

	if ((*end.path == '\0') && (end.vent->inode != 0))
		rest = chattr(end.vent->fs,end.vent->din,end.vent->inode,cmd,attr);
	else
	{
		rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,0,&inode);
		if (rest < 0)
			goto ERR;
		rest = chattr(start.vent->fs,start.vent->din,inode,cmd,attr);
		start.vent->fs->close(start.vent->din,inode);
	}
ERR:
	releaseVentObj(end.vent);

	return rest;
}


int sys_fchattr(int fdNum,int cmd,uint attr)
{
	int rest;
	F_DSC *fd;


	CALL_PRINT();

	if((fd=getFd(fdNum))==NULL)return -EBADF;
	if(fd->vent->type!=S_IFREG)return -EBADF;

	setVentObj(fd->vent);
	rest=chattr(fd->vent->fs,fd->vent->din,fd->vent->inode,cmd,attr);
	releaseVentObj(fd->vent);

	return rest;
}


/* ɲðꡣ */
int sys_ioctl(int fdNum,int cmd,int parm1,int parm2)
{
	F_DSC *fd;
	VENTRY *vent;


	CALL_PRINT();

	if ((fd = getFd(fdNum)) == NULL)
		return -EBADF;
	vent = fd->vent;

	return vent->fs->ioctl(vent->din,vent->inode,cmd,parm1,parm2);
}


int sys_link(const char *dst_path, const char *src_path)
{
	uint dst_inode;
	int din;
	int rest;
	SEARCH_PARAM dst_start,src_start,dst_end,src_end;


	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(dst_path))
		return -ENAMETOOLONG;
	if (PATH_MAX <= strlen(src_path))
		return -ENAMETOOLONG;

	/* dst_pathΥե򳫤 */
	dst_start.vent = getTopVdir(dst_path,&dst_start.path);
	if ((rest = getVentObj(&dst_start,&dst_end)) < 0)
		return rest;

	if (*dst_end.path == '\0')
	{
		if (dst_end.vent->type == S_IFDIR)
		{
			rest = -EISDIR;
			goto END;
		}
		din = dst_end.vent->din;
		dst_inode = dst_end.vent->inode;
	}
	else
	{
		din = dst_start.vent->din;
		if ((rest = dst_start.vent->fs->open(dst_start.path,din,dst_start.vent->inode,0,&dst_inode)) < 0)
			goto END;
	}

	/* src_pathΥե򸡺롣 */
	src_start.vent = getTopVdir(src_path,&src_start.path);
	if ((rest = getVentObj(&src_start,&src_end)) < 0)
		return rest;

	if (*src_end.path == '\0')
		rest = -EEXIST;
	else
	{
		if (src_start.vent->din != din)
			rest=-EXDEV;
		else
			rest = src_start.vent->fs->link(din,dst_inode,src_start.vent->inode,src_start.path);
	}
	releaseVentObj(src_end.vent);

	if (*dst_end.path != '\0')
		dst_start.vent->fs->close(din,dst_inode);
END:
	releaseVentObj(dst_end.vent);

	return rest;
}


int sys_symlink(const char *linkPath,const char *filePath)
{
	SEARCH_PARAM start,end;
	int rest = 0;
	
	
	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(filePath))
		return -ENAMETOOLONG;
	if (PATH_MAX <= strlen(linkPath))
		return -ENAMETOOLONG;

	if (*filePath == '\0')
		return -ENOENT;

	/* filePath򳫤 */
	start.vent = getTopVdir(filePath,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;

	if (*end.path == '\0')
	{
		rest = -EEXIST;
		goto END;
	}
	else
	{
		rest = start.vent->fs->symlink(start.path,start.vent->din,start.vent->inode,linkPath);
		if (rest < 0)
			goto END;
	}
END:
	releaseVentObj(end.vent);

	return rest;
}


int sys_readlink(const char *path,char *buf,size_t bufSize)
{
	uint inode;
	SEARCH_PARAM start,end;
	struct stat stat;
	int size;
	int rest = 0;


	CALL_PRINT();

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* ݥγǧ */
	if (checkMem(buf,bufSize) == -1)
		return -EFAULT;

	/* ե򸡺 */
	start.vent = getTopVdir(path,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;

	/* inode */
	if (*end.path == '\0')
		inode = end.vent->inode;
	else
		if ((rest = start.vent->fs->open(start.path,start.vent->din,start.vent->inode,0,&inode)) < 0)
			goto ERR;

	/* ե륿פǧ */
	rest = start.vent->fs->stat(start.vent->din,inode,&stat);
	if (rest < 0)
		goto ERR2;
	if ((stat.st_mode & S_IFLNK) == 0)
	{
		rest = -EINVAL;
		goto ERR2;
	}
	
	/* 󥯥ѥɤ */
	size = (stat.st_size <= bufSize)? stat.st_size : bufSize;
	rest = start.vent->fs->read(start.vent->din,inode,buf,size,0);

ERR2:	
	if (*end.path != '\0')
		start.vent->fs->close(start.vent->din,inode);
ERR:
	releaseVentObj(end.vent);

	return rest;
}


/*
 * PRIVATE
 * parameters : number of fds,fd_set,fdե饰polling flag
 * return : 0ʾ=Ѳfd number or -1=Ѳʤ or -2=fdʤ or error number
 */
STATIC int sendFdset(int nfds,fd_set *fdset,int flag,int polling)
{
	char *fdbit;
	F_DSC *fd;
	int rest = -2;
	int i,j,k;


	fdbit = (char*)fdset;
	for (i = 0; i < sizeof(fd_set); ++i){
		if (fdbit[i] != 0){
			for (j = 0; j < BYTE_BIT; ++j){
				if (fdbit[i] & (1 << j))
				{
					if ((k = i * BYTE_BIT + j) >= nfds){
						return rest;
					}

					if ((fd = getFd(k)) == NULL){
						return -EBADF;
					}
					rest = fd->vent->fs->select(fd->vent->inode,flag,polling);
					if (rest == 1){
						return k;
					}
					if (rest < 0){
						return rest;
					}
					rest = -1;
				}
			}
		}
	}

	return rest;
}

STATIC void INLINE clearFdset(fd_set **fdset,int cnt)
{
	int i;


	for(i=0;i<cnt;++i)
		fdset[i]->bitmap=0;
}

int sys_select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *errorfds, struct timeval *timeout)
{
	int seconds;
	int polling;
	int flag[3];
	fd_set *fdset[3];
	int i,cnt,cnt2,rest;

	CALL_PRINT();

	if (nfds < 0)
		return -EINVAL;

	if (timeout == NULL)
		/* seconds=-1 ʥȯޤǥ֥å롣 */
		seconds = -1;
	else
		seconds = timeout->tv_sec + (500000 <= timeout->tv_usec);

	cnt = 0;
	if (readfds != NULL)
	{
		if (checkMem(readfds,sizeof(fd_set)) == -1)
			return -EFAULT;
		flag[cnt] = FD_READ;
		fdset[cnt++] = readfds;
	}
	if (writefds != NULL)
	{
		if (checkMem(writefds,sizeof(fd_set)) == -1)
			return -EFAULT;
		flag[cnt] = FD_WRITE;
		fdset[cnt++] = writefds;
	}
	if (errorfds != NULL)
	{
		if (checkMem(errorfds,sizeof(fd_set)) == -1)
			return -EFAULT;
		flag[cnt] = FD_ERR;
		fdset[cnt++] = errorfds;
	}

	/* ʥԤ */
	if (cnt == 0)
	{
		sys_sleep(seconds);
		return 0;
	}

	/* polling뤫 */
	polling = (seconds == 0)? TRUE : FALSE;

	cnt2 = 0;
	for(;;)
	{
		for (i = 0; i < cnt; ++i)
		{
			rest = sendFdset(nfds, fdset[i], flag[i], polling);
			if (0 <= rest)							// Ѳ
			{
				clearFdset(fdset, cnt);
				((char*)fdset[i])[rest / BYTE_BIT] |= 1 << (rest % BYTE_BIT);

				return 1;
			}
			if (rest < -2)
				return rest;						// 顼
			if (rest == -2)
				cnt2 += 1;
		}

		if ((polling == TRUE) || (cnt2 == cnt))		/* polling or no fd. */
		{
			clearFdset(fdset,cnt);
			return 0;
		}

		seconds = sys_sleep(seconds);

		/* ʥߤ */
		doSignalRestart();
		if(isSigint() == TRUE){
			return -EINTR;
		}

		if (seconds == 0){
			polling = TRUE;
		}
	}
}


int sys_poll(struct pollfd *ufds, uint nfds, int miliSeconds)
{
	int polling,flag;
	int seconds;
	int i,rest;
	F_DSC *fd;


	CALL_PRINT();

	if (checkMem(ufds,sizeof(struct pollfd)) == -1){
		return -EFAULT;
	}

	polling = (miliSeconds == 0)? TRUE : FALSE;

	/* Init revents. */
	for (i = 0; i < nfds; ++i){
		ufds[i].revents = 0;
	}

	seconds = miliSeconds / 1000;
	for (;;){
		for (i = 0; i < nfds; ++i){
			/* Get file descriptor. */
			if (( fd = getFd(ufds[i].fd)) == NULL){
				ufds[i].revents = POLLNVAL;
				return 1;
			}

			/* Get flag. */
			switch(ufds[i].events){
				case POLLIN:
					flag = FD_READ;
					break;
				case POLLOUT:
					flag = FD_WRITE;
					break;
				case POLLERR:
				case POLLNVAL:
				case POLLHUP:
					flag = FD_ERR;
					break;
				default:
					return -EINVAL;
			}

			/* Send poll. */
			rest = fd->vent->fs->select(fd->vent->inode,flag,polling);
			if (rest == 1){
				ufds[i].revents=ufds[i].events;
				return 1;
			}
			else if (rest < 0){
				ufds[i].revents = POLLERR;
				return 1;
			}
		}

		if (polling == TRUE){
			break;
		}

		/* sleep */
		if (miliSeconds < 0){
			seconds = sys_sleep(-1);
		}
		else{
			seconds = sys_sleep(seconds);
		}

		/* ʥߤ */
		doSignalRestart();
		if(isSigint() == TRUE){
			return -EINTR;
		}

		if (seconds == 0){
			polling = TRUE;
		}
	}

	return 0;
}


int sys_getstatfs(struct statfs *buf,long bufsize,int flags)
{
	int cnt;


	CALL_PRINT();

	if (buf == NULL)
		cnt = INT_MAX;
	else cnt = bufsize / sizeof(struct statfs);
	
	return searchMount(buf,cnt);
}


int sys_statfs(const char *path, struct statfs *buf)
{
	int rest;
	SEARCH_PARAM start,end;
	VENTRY *vent,*p;


	CALL_PRINT();

	if (*path == '\0')
		return -ENOENT;

	/* ѥĹθ */
	if (PATH_MAX <= strlen(path))
		return -ENAMETOOLONG;

	/* Хåեå */
	if (checkMem(buf,sizeof(struct statfs)) == -1)
		return -EFAULT;

	/* ǥ쥯ȥõ */
	start.vent = getTopVdir(path,&start.path);
	if ((rest = getVentObj(&start,&end)) < 0)
		return rest;
	releaseVentObj(end.vent);

	if (*end.path == '\0')
	{
		if (end.vent->mount == NULL)
		{
			vent = end.vent;
			for (p = vent; p != rootEnt; p = p->parent)
				if (p->din == vent->din)
					break;
			memcpy(buf->f_mntonname,p->name,p->len);	/* directory on which mounted */
			buf->f_mntonname[p->len] = '\0';
		}
		else
		{
			vent = end.vent->mount;
			memcpy(buf->f_mntonname,vent->name,vent->len);	/* directory on which mounted */
			buf->f_mntonname[vent->len] = '\0';
		}
	}
	else
		return -ENOENT;

	return getFsstat(vent,buf);
}

/************************************* ̤ ***************************************/
int sys_lock()
{
	CALL_PRINT();

	return 0;
}


/************************************************************************************
 *
 * 
 *
 ************************************************************************************/


/*================================== ץ饤١ ======================================*/


static int setDevDir(const int devInode,FS *fs,const uint rootInode)
{
	uint inode;
	int rest;


	rest = fs->open("dev",devInode,rootInode,0,&inode);
	if (rest < 0)
		return rest;
	
	devEnt->din			= devInode;
	devEnt->fs			= fs;
	devEnt->inode		= inode;
	devEnt->refCount	+= 1;
	
	fs->close(devInode,inode);
	
	return 0;
}


/*=================================== Х =======================================*/


/*
 * Init filesystem
 */
void init_fs()
{
	/* rootȥ꡼ν */
	rootEnt->fs=      NULL;
	rootEnt->inode=   0;
	rootEnt->next=    rootEnt;
	rootEnt->prev=    rootEnt;
	rootEnt->parent=  rootEnt;
	rootEnt->child=   NULL;
	rootEnt->mount=   NULL;
	rootEnt->wait=    NULL;
	rootEnt->refCount=1;
	rootEnt->din=     0;
	rootEnt->type=    S_IFDIR;
	rootEnt->len=     1;
	rootEnt->name[0]= '/';

	/* devȥ꡼ν */
	devEnt->fs=      NULL;
	devEnt->inode=   0;
	devEnt->next=    devEnt;
	devEnt->prev=    devEnt;
	devEnt->parent=  devEnt;
	devEnt->child=   NULL;
	devEnt->mount=   NULL;
	devEnt->wait=    NULL;
	devEnt->refCount=1;
	devEnt->din=     0;
	devEnt->type=    S_IFDIR;
	devEnt->len=     3;
	devEnt->name[0]= 'd';
	devEnt->name[1]= 'e';
	devEnt->name[2]= 'v';

	/* Device filesystemȥ꡼ꡣ */
	rootEnt->child=devEnt;
}


/*
 * Mount device filesystem.
 * parameters : filesystem
 * return : 0 or error number
 */
int mount_dev_fs(FS *fs)
{
	return mountEntry(devEnt,fs,0,0);
}


/*
 * mount filesystem
 * parameters : device name,filesystem name
 * return : 0 or error number
 */
int mountRoot(const char *dev_path,const char *fs_name)
{
	uint inode;
	int devInode;
	FS *fs;
	int rest;


	/* Search file system number */
	if ((fs = search_fs(fs_name)) == NULL)
		return -ENODEV;

	/* ǥХץ */
	if ((devInode = mountDevice(dev_path)) < 0)
		return devInode;

	/* ޥȴؿƤӽФ */
	rest = fs->mount(devInode,&inode);
	if (rest < 0)
		goto ERR;

	/* 롼Ȳۥǥ쥯ȥꡣ */
	rootEnt->din = devInode;
	rootEnt->fs = fs;
	rootEnt->inode = inode;

	/* "dev"ǥ쥯ȥ */
	return setDevDir(devInode, fs, inode);

ERR:
	fs->close(0,devInode);

	return rest;
}


/*
 * ޥȥڡХååץǥХ
 * parameters : device name
 * return : device inode or error number
 */
int mountPageBackupDev(const char *dev_path)
{
	return mountDevice(dev_path);
}


/*
 * Init process0 file struct.
 * return 0 or error number
 */
int initProc1Fs()
{
	int rest;

	/*
	 * ɸϤ򳫤
	 */
	if((rest=sys_open("/dev/console",O_RDONLY,0))<0)return rest;	/* ɸϡ */
	if((rest=setFdFlag(rest,0))<0)return rest;						/* No exec fd close. */
	if((rest=sys_open("/dev/console",O_WRONLY,0))<0)return rest;	/* ɸϡ */
	if((rest=setFdFlag(rest,0))<0)return rest;
	if((rest=sys_open("/dev/console",O_WRONLY,0))<0)return rest;	/* ɸ२顼 */
	if((rest=setFdFlag(rest,0))<0)return rest;

	return 0;
}

/*************************************************************************************************/
int test_fs(uint value1, uint value2, uint value3, uint value4)
{
	printk("test_fs() 0x%x 0x%x 0x%x 0x%x\n", value1, value2, value3, value4);
	return 0;
}
/**************************************************************************************************/
