/*
 * fs.h
 *
 * Copyright 2002, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 */


#ifndef fs_h
#define fs_h


#include<config.h>
#include<types.h>
#include<wait.h>
#include<limits.h>
#include<sys/stat.h>
#include<share/fcntl.h>
#include<sys/unistd.h>
#include<share/select.h>
#include<share/stropts.h>
#include<sys/poll.h>


#ifndef ASM_FILE


#include<sys/mount.h>


enum{
	/* եǥץѥեμࡣ */
	FD_FORM_LOCALFILE=1,
	FD_FORM_DIRSTREAM,
	FD_FORM_NETSOKET,

	/* select()fd̡ */
	FD_READ= 1,	/* Read discriptor. */
	FD_WRITE=2,	/* Write discriptor. */
	FD_ERR=  3,	/* Except watch. */
};


/* fstatѥե¤Ρ */
typedef struct stat FSTAT;

/* ե륷ƥ।ե᡼¤ */
typedef struct FS{
	const char *name;
	/* prm:device inode,inode pointer ret:0 or error number */
	int (*mount)(int,uint*);
	/* prm:device inode ret:0 or error=-1 */
	int (*umount)(int,uint);
	/* prm:path,device inode,directory inode,flag,inode buffer ret:file format or error */
	int (*open)(const char*,int,uint,int,uint*);
	/* prm:device inode,inode ret:0 or error=-1 */
	int (*close)(int,uint);
	/* prm:device inode,inode,bffer,read bytes,begin byte ret:read bytes or error=-1 */
	int (*read)(int,uint,void*,size_t,size_t);
	/* prm:device param num,inode,bffer,write bytes,begin byte ret:write bytes or error=-1 */
	int (*write)(int,uint,void*,size_t,size_t);
	/* prm:device inode,inode,number of parameters,command,parameter ret:0 or error=-1 */
	int (*ioctl)(int,uint,int,...);
	/* prm:device inode,old dir inode,old path,new dir inode,new path ret:0 or error number */
	int (*rename)(int,uint,const char*,uint,const char*);
	/* prm:path,device inode,directory inode,mode,inode buffer ret:directory block or error=0 */
	int (*creat)(const char*,int,uint,int,uint*);
	/* prm:path,device inode,directory inode,inodebuffer,block,index buffer:0 or error number */
	int (*opendir)(const char*,int,uint,uint*,uint*,int*);
	/* prm:path,device inode,parent directory block,mode ret:directory block or error=0 */
	int (*mkdir)(const char*,int,uint,int);
	/* prm:device inode,inode,FSTAT buffer ret:0 or error number */
	int (*stat)(int,uint,struct stat*);
	/* prm:path,device inod,directory block,file type ret:0 or error number */
	int (*delete)(const char*,int,uint,int);
	/* prm:device inode,inode,block buffer,index buffer,name buffer ret:0 or error number */
	int (*readdir)(int,uint,uint*,int*,char*);
	/* prm:device inode,directory inode or inode,mode,uid,gid,access time,modified time ret:0 or error number */
	int (*chattr)(int,uint,uint,uint,uint,uint,uint);
	int (*link)(int,uint,uint,const char*);
	/* path,device inode,directory inode,link path:0 or error number */
	int (*symlink)(const char*,int,uint,const char*);
	/* param:inode,fdμ,polling flag  return:o=no or 1=yes */
	int (*select)(uint,int,int);
	/* param:device inode  return:0  */
	int (*statfs)(int din,struct statfs *buf);
	struct FS *next;
}FS;

/*
 * ۥȥ꡼
 * Note!
 *  fsinodeΰ֤ϥåȹ¤Τȥ󥯤ƤΤա
 */
typedef struct VENTRY{
	FS            *fs;			/* Filesystem. */
	uint           inode;		/* inode */
	struct VENTRY *next;		/* 饦ɥ󥯥ꥹȥͥ */
	struct VENTRY *prev;		/* 饦ɥ󥯥ꥹȥץӥ塼 */
	struct VENTRY *parent;		/* ƥȥ꡼ */
	struct VENTRY *child;		/* ΥΡɤΥȥåץȥ꡼ */
	struct VENTRY *mount;		/* ޥȥȥ꡼ */
	WAIT_QUEUE    *wait;		/* Lock wait. */
	ushort         refCount;	/* ȥ󥿡ޥȤ䥫ȥǥ쥯ȥ¹ԥե¹ԥ */
	uchar          din;			/* ǥХinode */
	uchar          len;			/* Name size. */
	int            type;		/* ե륿 */
	int            link;        /* ե󥯥 */
	char           name[0];		/* ѥ͡ */
}VENTRY;

/* եǥץ */
typedef struct{
	uchar  form;		/* ࡣ */
	uchar  refCount;	/* ȥȡ */
	ushort accessMode;	/* Access mode. */
	size_t offset;		/* ɤ߽񤭳ϥեåȡ */
	VENTRY *vent;		/* ۥȥ꡼ */
}F_DSC;

/*
 * ǥ쥯ȥꥹȥ꡼
 * աեǥץȶѤƻȤΤ"fd"ΰ֤ɬƬˤ
 */
typedef struct{
	F_DSC	fd;
	uint	block;		/* directory block. */
	int		index;		/* directory index. */
}DIR_STRAEM;

/* ץѥե빽¤ */
typedef struct{
	VENTRY *current_dir;	/* ȥǥ쥯ȥꡣ */
	F_DSC *fd[OPEN_MAX];
	uchar flag[OPEN_MAX];	/* եǥץե饰 */
	uchar fd_count;			/* եǥץץ */
	uchar next_fd;			/* ζեǥץ */
}FILE_STRUCT;

/* ե¹ѥեǥץ */
typedef struct{
	int    din;
	uint   inode;
	FS    *fs;
	size_t offset;		/* ɤ߽񤭳ϥեå */
}EXEC_FD;


extern FS *fs_info[];


extern void			 initExecFs();
extern void			 init_fs();
extern int			 mount_dev_fs(FS*);
extern int			 regist_fs(FS*);
extern FILE_STRUCT	*copyFileStruct(FILE_STRUCT*);
extern void			 releaseFileStruct(FILE_STRUCT*);
extern int			 init_file_struct(PROC*);
extern int			 setFd(VENTRY *vent,int form,int amode,int flag,F_DSC *fd);
extern F_DSC		*getFd(int);
extern int			 isFileOpenMax();
extern int			 initProc1Fs();
extern int			 get_inode(int);
extern int			 mountRoot(const char*,const char*);
extern int			 mountPageBackupDev(const char*);
extern int			 cmpString(const char *str1,const char *str2);
extern const char	*cmpPathNull(const char *path1,const char *path2);
extern int			 cmpPath(const char *name,const char *path,int count);
extern size_t		 pathLen(const char *path);
extern int			 isLastPath(const char *path);

extern int			 exec_open(const char*,EXEC_FD*);
extern int			 exec_read(EXEC_FD*,void*,size_t);
extern int			 exec_close(EXEC_FD*);
extern int			 sys_open();
extern int			 sys_close();
extern int			 sys_read();
extern int			 sys_write();
extern int			 sys_ioctr();
extern int			 sys_lseek();
extern int			 sys_mkdir();
extern int			 sys_creat();
extern int			 sys_rename();
extern int			 sys_unlink();
extern int			 sys_rmdir();
extern int			 sys_lock();
extern int			 sys_mount();
extern int			 sys_umount();
extern int			 sys_opendir();
extern int			 sys_closedir();
extern int			 sys_readdir();
extern int			 sys_rewinddir();
extern int			 sys_chdir();
extern int			 sys_fchdir();
extern int			 sys_getcwd();
extern int			 sys_stat();
extern int			 sys_chattr();
extern int			 sys_fchattr();
extern int			 sys_umount_root();
extern int			 sys_fcntl();
extern int			 sys_ioctl();
extern int			 sys_link();
extern int			 sys_symlink();
extern int			 sys_readlink();
extern int			 sys_select();
extern int			 sys_poll();
extern int			 sys_getstatfs();
extern int			 sys_statfs();



#endif


#endif
